#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <cstdarg>
#include <ctime>
#include <iostream>
#include <fstream>
#include "opengl.h"
#include "soil/SOIL.h"
#include "config.h"

#include "vmath.h"
#include "Camera.h"
#include "defines.h"
#include "Mesh.h"

#include "ParticleSystem.h"
#include "Wind.h"
#include "WindLBM.h"
#include "Terrain.h"
#include "Snow.h"
#include "CudaHelpers.h"

#include "Road.h"
#include "Simulator.h"

//to enable vertical sync
#ifdef WIN32
#include <windows.h>
#include "wglext.h"
#include <string.h>
#endif

#if !(NO_QUAD_BUFFER == 1)
#define QUAD_STEREO
#endif

using namespace std;

Simulator* pEngine = 0;

static float deltaTime;

typedef unsigned char uchar;
typedef unsigned short ushort;
typedef unsigned int uint;

int mouseX, mouseY;
Camera camera;
// Default value of what to render
static bool drawWindField = false;
static bool drawObstacleField = false;
static bool drawPressureField = false;
static bool drawTerrain = true;
static bool drawMeshes = false;
bool stepMode = false;
int stepCount = 0;
float windLimit = 60.0f;

// Scene lighting parameters

    /* Not used?
   float lightAmbient[] = { 0.5f, 0.5f, 0.5f, 1.0f };
   float lightDiffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
   float lightPosition[] = { 64.0f, 10.0f, 128.0f, 1.0f };
 */

static int numFrames = 0;
// Vectors used to store statistical values
std::vector<float> fpsEntries;
std::vector<float> mlupsEntries;

// Used for rendering
Shader stereoShader;

GLuint fbo;
GLuint depthbuffer;
GLuint target1;
GLuint target2;

/**
 * "Die" function, called when exiting unexpectedly
 */
void die(const char *fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
    exit(1);
}


/**
 * Initilize the stencil mask used to render interlaced stereoscopic 3d
 */
void initStencilMask() {
/*
    glClearStencil(0);
    glClear(GL_STENCIL_BUFFER_BIT);
    glStencilFunc(GL_ALWAYS, GL_REPLACE, GL_REPLACE);
    glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);

    glEnable(GL_STENCIL_TEST);
    glEnable2D();
    glBegin(GL_LINES);
    for(int i = 0; i < conf.screenHeight/2; ++i) {
        glVertex2i(0, i*2);
        glVertex2i(conf.screenWidth-1, i*2);
    }
    glEnd();
    glDisable2D();
    glDisable(GL_STENCIL_TEST);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    */
} 

/**
 * Initialize rendering
 */
void Init() {
    glShadeModel(GL_SMOOTH);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    // Set the background color
    glClearColor((float)((conf.clearColor >> 24) & 0xff)/255.0f,
            (float)((conf.clearColor >> 16) & 0xff)/255.0f,
            (float)((conf.clearColor >>  8) & 0xff)/255.0f,
            (float)((conf.clearColor >>  0) & 0xff)/255.0f);
    std::cout<<"background color is: "<<((float)((conf.clearColor >> 24) & 0xff)/255.0f)<<" , "<<((float)((conf.clearColor >> 16) & 0xff)/255.0f)<<" , "<<((float)((conf.clearColor >> 8) & 0xff)/255.0f)<<" , "<<((float)((conf.clearColor >> 0) & 0xff)/255.0f)<<std::endl;
    glClearDepth(1.0f);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);

    // Enable backface culling, mainly for the terrain
    // Can gain some extra rendering performance, but does not look good on a interactive demo
    //glEnable(GL_CULL_FACE);

    glEnable(GL_POINT_SMOOTH);

    initStencilMask();

    glGenTextures(1, &target1);
    glBindTexture(GL_TEXTURE_2D, target1);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, conf.screenWidth, conf.screenHeight/2, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glBindTexture(GL_TEXTURE_2D, 0);

    glGenTextures(1, &target2);
    glBindTexture(GL_TEXTURE_2D, target2);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, conf.screenWidth, conf.screenHeight/2, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glBindTexture(GL_TEXTURE_2D, 0);

    glGenRenderbuffersEXT(1, &depthbuffer);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depthbuffer);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT, conf.screenWidth, conf.screenHeight/2);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);

    glGenFramebuffersEXT(1, &fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, target1, 0);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D, target2, 0);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, depthbuffer);

    GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    if(status != GL_FRAMEBUFFER_COMPLETE_EXT)
        die("fbo error\n");
    
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    stereoShader.load("stereo.vert", "stereo.frag");
    glCheckError("Failed to start stereo");
}

/**
 * Render the scene, called by the display function
 */
void display(Snow &snow, Terrain &terrain, Wind &wind, WindLBM &windLBM, vector<Mesh*>& meshes) {
    glLoadIdentity();
    if (!conf.rotate)
        camera.look();
    else
    {
        static float rot = 0;
        rot += 0.005;
        float x = conf.rotateX + conf.rotateRadius*sin(rot),
              y = conf.rotateHeight,
              z = conf.rotateZ + conf.rotateRadius*cos(rot);
        camera.pos.x = x; camera.pos.y = y; camera.pos.z = z;
        gluLookAt(x, y, z, conf.rotateX, conf.rotateY, conf.rotateZ, 0, 1, 0);
    }

    if(drawTerrain) {
        glPushMatrix();
        //terrain.Render();
        glPopMatrix();
    }
/*
    if (drawMeshes)
    {
        for (size_t i = 0; i < meshes.size(); i++)
        {
            glPushMatrix();
            meshes[i]->render();
            glPopMatrix();
        }
    }
    */

    /*
    if(conf.renderSnow) {
        snow.Render(&camera);
    }
    */

#ifdef PRINT_DEBUG
    std::cout << "Drawing: Rendering fields." << std::endl;
#endif

    // Debug rendering
    if(drawObstacleField || drawPressureField || drawWindField) {
        glPushMatrix();
        glScalef(SCENE_X, SCENE_Y, SCENE_Z);

        glEnable(GL_POINT_SPRITE);
        glPointSize(10.0f);

        if(drawObstacleField) {
            if(conf.usingLBM)
                windLBM.RenderObstacles();
            else
                wind.RenderObstacles();
        }

        // Dont render pressure field if using the LBM-simulation
        if(drawPressureField) {
            if(conf.usingLBM)
                windLBM.RenderPressure();
            else
                wind.RenderPressure();
        }

        glDisable(GL_POINT_SPRITE);

        if(drawWindField) {
            if(conf.usingLBM)
                windLBM.RenderDebug(windLimit);
            else
                wind.RenderDebug(windLimit);
        }

        glPopMatrix();
    }
}

/**
 * Renders to the screen, either regular or stereoscopic mode
 * Both quad-buffered and interlaced stereo
 */
void displayStereo(Snow &snow, Terrain &terrain, Wind &wind, WindLBM &windLBM, vector<Mesh*>& meshes) {
    if(conf.stereo) {
#ifdef QUAD_STEREO
        glDrawBuffer(GL_BACK_LEFT);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        camera.strafe(-conf.eyeSeparation*0.5f);
        display(snow, terrain, wind, windLBM, meshes);

        glDrawBuffer(GL_BACK_RIGHT);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        camera.strafe(conf.eyeSeparation);
        display(snow, terrain, wind, windLBM, meshes);

        camera.strafe(-conf.eyeSeparation*0.5f);
#else
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
        glPushAttrib(GL_VIEWPORT_BIT);
        glViewport(0, 0, conf.screenWidth, conf.screenHeight/2);

        glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        camera.strafe(-conf.eyeSeparation*0.5f);
        display(snow, terrain, wind, windLBM, meshes);
        glDrawBuffer(GL_COLOR_ATTACHMENT1_EXT);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        camera.strafe(conf.eyeSeparation);
        display(snow, terrain, wind, windLBM, meshes);
        camera.strafe(-conf.eyeSeparation*0.5f);

        glPopAttrib();
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

#define SHADER_FOO 0
#ifdef SHADER_FOO
        stereoShader.bind();
        stereoShader.uniform("texture1", 0);
        stereoShader.uniform("texture2", 1);

        glActiveTexture(GL_TEXTURE0);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, target1);
        glActiveTexture(GL_TEXTURE1);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, target2);

        glEnable2D();
        /*
        glBegin(GL_QUADS);
        glTexCoord2i(0, 0); glVertex2i(0, 0);
        glTexCoord2i(1, 0); glVertex2i(conf.screenWidth, 0);
        glTexCoord2i(1, 1); glVertex2i(conf.screenWidth, conf.screenHeight);
        glTexCoord2i(0, 1); glVertex2i(0, conf.screenHeight);
        glEnd();
        */
        glDisable2D();

        glActiveTexture(GL_TEXTURE0);
        glDisable(GL_TEXTURE_2D);
        glActiveTexture(GL_TEXTURE1);
        glDisable(GL_TEXTURE_2D);
        stereoShader.unbind();
#else
        glEnable(GL_STENCIL_TEST);
        glStencilFunc(GL_NOTEQUAL, 1, 1);
        drawFullscreenQuad(target1);
        glStencilFunc(GL_EQUAL, 1, 1);
        drawFullscreenQuad(target2);
        glDisable(GL_STENCIL_TEST);
#endif

#endif
    }
    else {
        glDrawBuffer(GL_BACK);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        display(snow, terrain, wind, windLBM, meshes);
    }
}

/**
 * Render textures for polarized stereoscopic rendering
 */
void drawFullscreenQuad(GLuint texture) {
/*
    glActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, texture);
    glMatrixMode (GL_MODELVIEW); glPushMatrix (); glLoadIdentity (); glMatrixMode (GL_PROJECTION); glPushMatrix (); glLoadIdentity ();
    glBegin (GL_QUADS); 
    glTexCoord2i(0, 0); glVertex3i (-1, -1, -1);
    glTexCoord2i(1, 0); glVertex3i (1, -1, -1);
    glTexCoord2i(1, 1); glVertex3i (1, 1, -1);
    glTexCoord2i(0, 1); glVertex3i (-1, 1, -1);
    glEnd ();
    glPopMatrix (); glMatrixMode (GL_MODELVIEW); glPopMatrix ();
    */
}

/**
 * Reshape the frustum when the size of the window changes, callback function.
 */
void GLFWCALL reshape(int width, int height) {
    
    glViewport(0, 0, width, height);
/*
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0f, (float)width / (float)height, 2.0f, 1000000.0f);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    */
    pEngine->resize(width, height);
}

/**
 * Handle keyboard input, callback function.
 * Non-toggle keys are handled inside the main loop (depend on time)
 */
void GLFWCALL keyHandler(int key, int action) {
    if(action == GLFW_PRESS) {
        pEngine->processKeyEvents(key);
        switch(key) {
            case 'T':
                conf.stereo = !conf.stereo;
                break;
            case 'F':
                drawWindField = !drawWindField;
                break;
            case 'O':
                drawObstacleField = !drawObstacleField;
                break;
            case 'P':
                drawPressureField = !drawPressureField;
                break;
            case 'R':
                drawTerrain = !drawTerrain;
                break;
            case 'H':
                drawMeshes = !drawMeshes;
                break;
            case '.':
                conf.rotate = !conf.rotate;
                break;
            case 'M':
                conf.simulateWind = !conf.simulateWind;
                printf("simulateWind is now: %d\n", conf.simulateWind);
                break;
            case 'E':
                conf.simulateSnow = !conf.simulateSnow;
                break;
            case 'N':
                conf.renderSnow = !conf.renderSnow;
                break;
            case 'C':
                stepMode = !stepMode;
                break;
            case '1':
                conf.enableDisplay = !conf.enableDisplay;
                break;
            case GLFW_KEY_ENTER:
                if(stepMode)
                    ++stepCount;
                break;
                
        }
    }
}

/**
 * Handle mouse motion events (move the camera), callback function.
 */
void GLFWCALL mouseMotion(int x, int y) {
    int dx = x - mouseX;
    int dy = y - mouseY;

    mouseX = x;
    mouseY = y;

    // Comment to make sure that the camera is not moved by accident
    camera.turn(-conf.mouseSensitivityX*dx, conf.mouseSensitivityY*dy);
    pEngine->mouseMove(dx, -dy);
}

/**
 * Handle mouseclick events, callback function.
 */
void GLFWCALL mouseButton(int button, int action) {
}

/**
 * Main function. See debug printouts for information about each section
 */
int main(int argc, char *argv[]) {
    double last_time = 0;
    float timer = 0;
    int frame_count = 0;
    bool running;
    int seconds = 0;
    float temp_mlups = 0;

     
    std::cout << "Start snow simulation application\n";

#ifdef WIN32
    srand((long)time(NULL));
#else
    srand48( (long)time(NULL) );
#endif
    // Use for predictable debugging
    //srand(2);

    // Load configuration file 
    load_config("config.txt");

    
    glfwInit();
    /*
#ifdef QUAD_STEREO
    glfwOpenWindowHint(GLFW_STEREO, GL_TRUE);
#endif
    glfwOpenWindowHint( GLFW_FSAA_SAMPLES, 4 );
    glfwOpenWindowHint( GLFW_OPENGL_VERSION_MAJOR, 3 );
    glfwOpenWindowHint( GLFW_OPENGL_VERSION_MINOR, 2 );
    glfwOpenWindowHint( GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE );
*/
    if(!glfwOpenWindow(conf.screenWidth, conf.screenHeight, 8, 8, 8, 8, 24, 0, GLFW_WINDOW)) {
        die("Failed to open window\n");
    }

    if (conf.usingLBM)
      glfwSetWindowTitle("Snow simulation by IDI/NTNU HPC lab (LBM)");
    glfwSetWindowTitle("Snow simulation by IDI/NTNU HPC lab");

    glfwOpenWindowHint(GLFW_WINDOW_NO_RESIZE, GL_TRUE);
    //glfwOpenWindowHint(GLFW_FSAA_SAMPLES, 4);

    glfwEnable(GLFW_STICKY_KEYS);
    // Comment out for debugging without locking the mouse to the window
    glfwDisable(GLFW_MOUSE_CURSOR);

   

    GLenum err = glewInit();
    if(GLEW_OK != err)
        die("GLEW initialization error: %s\n", glewGetErrorString(err));


    Init();
    
    CUDA_Init(conf.device);
    glCheckError("Failed CUDA INIT");
    std::cout << "Cuda initialized\n\n";

    camera.init(conf.camX, conf.camY, conf.camZ, conf.camYaw, conf.camPitch);
    std::cout << "Rendering window initialized\n\n";

    WindowProperties properties = {conf.screenWidth, conf.screenHeight};
    pEngine = new Simulator(properties);
    
    glfwSetWindowSizeCallback(reshape);
    glfwSetKeyCallback(keyHandler);
    glfwSetMousePosCallback(mouseMotion);
    glfwSetMouseButtonCallback(mouseButton);

/*
    Terrain terrain;
    vector<Mesh*> meshes;
    Snow snow;
    Wind wind;
    WindLBM windLBM;

    Road road(conf.roadDefinitionFile, SCENE_X/float(conf.map.width*conf.map.resX), conf.map.minHeight);
    meshes.push_back(&(road.getMesh()));

    terrain.Init(conf.map.heightmap, conf.map.normalmap, conf.map.width, conf.map.height, road);
    std::cout << "Terrain loaded (with " << conf.map.width << "x" << conf.map.height << " heightmap)" << "\n";
    std::cout << "Heightmap loaded from " << conf.map.heightmap << ", normalmap from " << conf.map.normalmap << "\n\n";

    std::cout << "Wind field initialization started";
    if(conf.usingLBM)
        std::cout << " (LBM simulation)";
    else
        std::cout << " (finite difference Navier-Stokes)";
    std::cout << "\n\n";

    if(conf.usingLBM)
        windLBM.Init(conf.windX, conf.windY, conf.windZ, conf.map.width, conf.LBMOmega, conf.LBMViscosity, conf.LBMVelFactor, conf.LBMInitVel, conf.LBMInitRho, conf.LBMBoundaryVelFactor);
    else
        wind.Init(conf.windX, conf.windY, conf.windZ, conf.map.width);

    snow.Init(conf.numParticles);
    std::cout << "Snow particles initialized (" << conf.numParticles << " particles)\n\n";

    std::cout << "Wind field initialization successful\n\n";

    std::cout << "Updating obstacles\n";
    
    for(int i = 0; i < 7; ++i) {
        for(int y = 0; y < layers[i]; ++y)
            snow.UpdateObstacles(terrain.gvbo, i, y);
    }
*/
    int layers[] = { 1, 1, conf.windY-1, 1, conf.windY-2, 1, 1 };
    pEngine->init();
    std::ofstream resfile("results.txt");

    std::cout << "Entering main loop\n";

    float obs_timer = 0;
    int obs_stage = 0;
    int obs_y = 0;

    // Main loop
    do {
#ifdef PRINT_DEBUG
        std::cout << "Frame: Starting to paint frame num: " << numFrames + 1 << std::endl;
#endif

        if(last_time == 0)
            last_time = glfwGetTime();
        double new_time = glfwGetTime();
        deltaTime = (float)(new_time - last_time);
        last_time = new_time;

        timer += deltaTime;
        ++frame_count;
        if(timer > 1.0f) {
            float fps = (float)frame_count/timer;
            std::cout << "fps: " << fps << std::endl;
            fpsEntries.push_back(fps);
            resfile << fps << std::endl;
            timer = 0;
            frame_count = 0;
            if(++seconds == conf.timeLimit)
                break;
        }

        obs_timer += deltaTime;

        if(obs_timer > 0.1f) {
            obs_timer = 0;
            pEngine->UpdateObstacles(obs_stage, obs_y++);
            if(obs_y >= layers[obs_stage]) {
                obs_y = 0;
                ++obs_stage;
                if(obs_stage > 6)
                    obs_stage = 0;
            }
        }

        if(conf.frameLimit && numFrames > conf.frameLimit)
            break;

        double realSpeed = conf.movementSpeed * deltaTime;
        if (glfwGetKey(GLFW_KEY_LSHIFT))
            realSpeed *= 4.;
        
        if (glfwGetKey(GLFW_KEY_LCTRL))
            realSpeed /= 8.;

        if(glfwGetKey('W')){
            pEngine->step(realSpeed);
            camera.step(realSpeed);
        }
        if(glfwGetKey('S')){
            pEngine->step(-realSpeed);
            camera.step(-realSpeed);
        }
        
        if(glfwGetKey('A')){
         pEngine->strafe(-realSpeed);
         camera.strafe(-realSpeed);
        }
        if(glfwGetKey('D')){
          pEngine->strafe(realSpeed);
         camera.strafe(realSpeed);
         }
        if(glfwGetKey(GLFW_KEY_SPACE))
        { 
            pEngine->move(0,realSpeed,0);
            camera.move(0,realSpeed,0);
        }
        if(glfwGetKey(',')) windLimit -= 2.0f * deltaTime;
        if(glfwGetKey('.')) windLimit += 2.0f * deltaTime;

        if(glfwGetKey('C'))
            conf.eyeSeparation += 1.0f * deltaTime;
        if(glfwGetKey('V'))
            conf.eyeSeparation -= 1.0f * deltaTime;

        if(glfwGetKey('Z'))
        {
        /*
            if(conf.usingLBM) 
                windLBM.Init(conf.windX, conf.windY, conf.windZ, conf.map.width, conf.LBMOmega, conf.LBMViscosity, conf.LBMVelFactor, conf.LBMInitVel, conf.LBMInitRho, conf.LBMBoundaryVelFactor);
            else 
                wind.Init(conf.windX, conf.windY, conf.windZ, conf.map.width);
            snow.Init(conf.numParticles);
            terrain.Init(conf.map.heightmap, conf.map.normalmap, conf.map.width, conf.map.height, road);
            */
        }
        if(glfwGetKey('X'))
        {
            //terrain.Init(conf.map.heightmap, conf.map.normalmap, conf.map.width, conf.map.height, road);
        }

#ifdef PRINT_DEBUG
        std::cout << "Frame: Checks for timers and keys are done." << std::endl;
#endif

        pEngine->update(deltaTime);

#ifdef PRINT_DEBUG	
        std::cout << "Frame: Done simulating windfield." << std::endl;
#endif

#ifdef PRINT_DEBUG
        std::cout << "Frame: Done simulating particles." << std::endl;
#endif
        if(conf.enableDisplay)
            pEngine->draw(camera);
            //displayStereo(snow, terrain, wind, windLBM, meshes);

#ifdef PRINT_DEBUG
        std::cout << "Frame: Done drawing." << std::endl;       
#endif

        if(conf.enableDisplay)
            glfwSwapBuffers();
        // Poll for events, if it is not called via swapbuffers
        else
            glfwPollEvents();

        running = !glfwGetKey(GLFW_KEY_ESC) && glfwGetWindowParam(GLFW_OPENED);

        numFrames++;

#ifdef PRINT_DEBUG
        std::cout << "Frame: Number of frames now painted: " << numFrames << std::endl;
#endif
    } while(running);

    printf("Jumped out of the main loop.\n");

    // Print the average FPS before quitting
    float fpsSum = 0.0;
    std::vector<float>::iterator iter = fpsEntries.begin();

    // Skip the first value
    if(iter != fpsEntries.end()) iter++;
    while(iter != fpsEntries.end()) {
        fpsSum += *iter;	
        iter++;
    }
    std::cout << "Average fps in this run was: " << fpsSum / (fpsEntries.size()-1) << std::endl;

#ifdef COUNT_MLUPS
    // Print the average MLUPS before quitting
    float mlupsSum = 0.0;
    std::vector<float>::iterator miter = mlupsEntries.begin();

    // Skip the first value
    if(miter != mlupsEntries.end()) miter++;
    while(miter != mlupsEntries.end()) {
        mlupsSum += *miter;	
        miter++;
    }
    std::cout << "Average MLUPS in this run was: " << mlupsSum / (mlupsEntries.size()-1) << std::endl;
#endif

    glfwTerminate();

    printf("GLFW terminated, returning 0.\n");

    // explicitly calling the wind destructor
    //windLBM.~WindLBM();
    //windLBM.has_allocated = false;
    //delete [] windLBM;

    // To stop the command window from closing
    cin.get();

    return 0;
}

