#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include "Mesh.h"
#include "soil/SOIL.h"
#include "utils.h"

using namespace std;

bool loadMtlLib(string filename, MaterialCollection& mats)
{
    cout << "Reading " << filename.c_str() << endl;
    ifstream mtlfile(filename.c_str());
    if (!mtlfile.is_open())
        return false;



    string matname;
    while (!mtlfile.eof() && !mtlfile.fail())
    {
        string buffer;
        getline(mtlfile, buffer);

        string key;
        stringstream ss(buffer);
        ss >> key;
        if (key[0] == '#' || key == "")
            continue;

        if (key == "newmtl")
        {
            ss >> matname;
            mats.materials[matname] = Material();
        }
        else if (key == "Ka")
        {
            Material& mat = mats.materials[matname];
            ss >> mat.ka.x >> mat.ka.y >> mat.ka.z >> mat.ks.w;
        }
        else if (key == "Kd")
        {
            Material& mat = mats.materials[matname];
            ss >> mat.kd.x >> mat.kd.y >> mat.kd.z >> mat.ks.w;
        }
        else if (key == "Ks")
        {
            Material& mat = mats.materials[matname];
            ss >> mat.ks.x >> mat.ks.y >> mat.ks.z >> mat.ks.w;
        }
        else if (key == "map_Kd")
        {
            Material& mat = mats.materials[matname];
            string texfile;
            ss >> texfile;
            mat.texture = SOIL_load_OGL_texture(texfile.c_str(), SOIL_LOAD_AUTO, SOIL_CREATE_NEW_ID, SOIL_FLAG_MIPMAPS | SOIL_FLAG_TEXTURE_REPEATS);
//            printf("Material texture: %d From file: %s\n", mat.texture, texfile.c_str());
        }
        else
        {
            printf("\nWarning: Key %s is not supported in mtllib.\n", key.c_str());
        }
    }
    mtlfile.close();

    return true;
}

Mesh::Mesh()
{
}

void trim_beginning(char*& s)
{
    while (*s == '\n' || *s == '\r' || *s == ' ')
    {
        s++;
    }
}

char* next_n(char* s, int n)
{

    char* tmp = new char[n+1];
    tmp[n] = 0;
    for (int i = 0; i < n; i++)
        tmp[i] = s[i];

    return tmp;
}

float read_float(char*& s)
{
    float sign = 1,
          ans = 0;
    size_t before = 0, after = 0; 

    trim_beginning(s);

    if (*s == '-')
    {
        sign = -1;
        s++;
    }

    //Parse the stuff before the decimal point
    while (*s >= '0' && *s <= '9')
    {
        before = before*10 + (*s - '0');
        s++;
    }
    ans += before;

    //Parse the stuff after the decimal point
    if (*s == '.')
    {
        s++;
        size_t n = 0;
        while (*s >= '0' && *s <= '9')
        {
            after = after*10 + (*s - '0');
            s++;
            n++;
        }
        ans += after / pow(10., n);
    }
    if (*s == 'e')
    {
        while (*s != ' ' && *s != '\n')
            s++;
    }


    return ans * sign;
}

int read_positive_int(char*& s)
{
    int ans = 0;

    trim_beginning(s);

    while (*s >= '0' && *s <= '9')
    {
        ans = ans*10 + (*s - '0');
        s++;
    }

    return ans;
}

//Mesh loader; it's not very flexible on the format due to performance. All faces MUST have both normals, vertices and texture coordinates.
bool Mesh::loadObj(string filename, float scale, bool flipYZ, float translateY)
{
    ifstream objfile(filename.c_str());
    if (!objfile.is_open())
        return false;

    char* file_content = 0;
    int length = objfile.seekg(0, ios::end).tellg();
    objfile.seekg(0,ios::beg);
    file_content = new char[length+1];
    file_content[length] = 0;
    objfile.read(file_content, length);
    objfile.close();
    char* f = file_content;

    while (f - file_content < length)
    {
        //To remove empty lines and stuff like that
        trim_beginning(f);
        if (f - file_content >= length)
            break;
//        printf("%c%c\n", *f, *(f+1)); fflush(stdout);

        //Skip comments and stuff like that.
        if (*f == '#' || *f == 'g' || *f == 'o')
        {
            while (f - file_content < length && *f != '\n' )
            {
//                printf("%c\n", *f); fflush(stdout);

                ++f;
            }
//            printf("%c\n", *f); fflush(stdout);
//            ++f;
            continue;
        }

        if (*f == 'v' && f[1] == ' ')
        {
            f++;
            float v1, v2, v3;
            v1 = read_float(f);
            v2 = read_float(f);
            v3 = read_float(f);
            if (flipYZ) swap(v2,v3);
            v2 += translateY;
            
            vertices.push_back(Vec3((v1)*scale,(v2)*scale,(v3)*scale));
        }
        //mtllib
        else if (*f == 'm')
        {
            while (*f != ' ')
                f++;

            trim_beginning(f);
            char* fn = new char[255];
            char* fp = fn;
            

            while (*f != '\n' && *f != '\r')
            {
                *fp = *f;
                fp++;
                f++;
            }
            *fp = 0;

            if (!loadMtlLib(string(fn), materials))
                fprintf(stderr, "WARNING: Could not load material library %s.\n", fn);

            delete[] fn;
        }
        //vn
        else if (f[0] == 'v' && f[1] == 'n')
        {
            f += 2;
            float v1, v2, v3;
            v1 = read_float(f);
            v2 = read_float(f);
            v3 = read_float(f);
            normals.push_back(Vec3(v1,v2,v3));
//            printf("%f, %f, %f\n", v1, v2, v3);
        }
        //vt
        else if (f[0] == 'v' && f[1] == 't')
        {
            f += 2;
            float v1, v2;
            v1 = read_float(f);
            v2 = read_float(f);
            texcoords.push_back(Vec2(v1,v2));
//            printf("%f, %f\n", v1, v2);
        }
        else if (*f == 'f' && f[1] == ' ')
        {
            f++;

            //New face
            vIndices.push_back(vector<int>());
            vnIndices.push_back(vector<int>());
            vtIndices.push_back(vector<int>());

            vector<int>& v = vIndices.back();
            vector<int>& vn = vnIndices.back();
            vector<int>& vt = vtIndices.back();

            for (int ii = 0; ii < 3; ii++)
            {
                //Need to split the string on slash
                trim_beginning(f);

                int vi = -1, vni = -1, vti = -1;
                int j = 0; 
                vi = read_positive_int(f)-1; f++;                
                vti = read_positive_int(f)-1; f++;
                vni = read_positive_int(f)-1; 

//                size_t slashpos[2] = {tmp.length(), tmp.length()};
//                for (size_t i = 0; *f != ; i++)
//                {
//                    if (tmp[i] == '/')
//                    {
//                        slashpos[j] = i;
//                        j++;
//                    }
//                }
                
//                vi = atoi(tmp.substr(0, slashpos[0]).c_str()) - 1;
//                if (slashpos[0] < tmp.length())
//                    vti = atoi(tmp.substr(slashpos[0]+1, slashpos[1]-slashpos[0]).c_str()) - 1;
//                if (slashpos[1] < tmp.length())
//                    vni = atoi(tmp.substr(slashpos[1]+1, tmp.length() - slashpos[1]).c_str()) - 1;

                v.push_back(vi);
                vt.push_back(vti);
                vn.push_back(vni);
            }
        }
        //usemtl
        else if (*f == 'u')
        {
            while (*f != ' ')
                f++;

            trim_beginning(f);
            char* mn = new char[255];
            char* mp = mn;
            

            while (*f != '\n' && *f != '\r')
            {
                *mp = *f;
                mp++;
                f++;
            }
            *mp = 0;



//            string matname;
//            ss >> matname;
            materials.matStart.push_back(pair<int, string>(vIndices.size(), string(mn)));
            printf("Using material %s at index %d.\n", mn, vIndices.size());

            delete[] mn;
        }
        else
        {
            char* nextn = next_n(f-10, 20);
            printf("Key is not supported in obj file. Prev and next 10 characters: %s\n", nextn);
            delete[] nextn;

            return false;
        }

    }
    return true;
}

void Mesh::initBuffers()
{
	// Bind normalmap to a texture
//    glGenTextures(1, &ntex);
//    glBindTexture(GL_TEXTURE_2D, ntex);
//    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
//    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
//    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
//    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
//    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width-1, height-1, 0, GL_RGB, GL_UNSIGNED_BYTE, nmap);
}

void Mesh::render()
{
    static bool printed = false;
    glEnable(GL_TEXTURE_2D);

    glColor4f(1,1,1,1);

    size_t currentMat = 0;
    glBegin(GL_TRIANGLES);
    for (size_t i = 0; i < vIndices.size(); i++)
    {
        if (currentMat < materials.matStart.size() && i >= materials.matStart[currentMat].first)
        {
            Material& mat = materials.materials[materials.matStart[currentMat].second];
            glEnd();
            glColor4f(mat.kd.x, mat.kd.y, mat.kd.z, mat.kd.w);
            glBindTexture(GL_TEXTURE_2D, mat.texture);
            glBegin(GL_TRIANGLES);
            currentMat++;
        }
        vector<int>& vi = vIndices[i],
                   & vn = vnIndices[i],
                   & vt = vnIndices[i];

        glNormal3f(normals[vn[0]].x, normals[vn[0]].y, normals[vn[0]].z);
        glTexCoord2f(normals[vt[0]].x, normals[vt[0]].y);
        glVertex3f(vertices[vi[0]].x, vertices[vi[0]].y, vertices[vi[0]].z);

        glNormal3f(normals[vn[1]].x, normals[vn[1]].y, normals[vn[1]].z);
        glTexCoord2f(normals[vt[1]].x, normals[vt[1]].y);
        glVertex3f(vertices[vi[1]].x, vertices[vi[1]].y, vertices[vi[1]].z);

        glNormal3f(normals[vn[2]].x, normals[vn[2]].y, normals[vn[2]].z);
        glTexCoord2f(normals[vt[2]].x, normals[vt[2]].y);
        glVertex3f(vertices[vi[2]].x, vertices[vi[2]].y, vertices[vi[2]].z);
    }
    glEnd();
    printed = true;
    glColor4f(1,1,1,1);
}
