#include "CudaIncludes.cuh"
#include "utils.h"
#include "ParticleSystem.cuh"
#include "SnowSystemKernels.cu"

//#include "utils.h"
//#include "SnowSystem.cuh"

//#include <cuda_gl_interop.h>


// Particle arrays
float4 *part_vel;		// velocity(x,y,z) and Vmax,z(w)
float  *part_rot_radius;	// rotation radius
float  *part_rot_omega;		// angular velocity

static float calc_omega() {
        float rotOmega = float(myrandom() * (M_PI / 3 - M_PI / 4) + M_PI / 4);
        if (myrandom() < 0.5) rotOmega *= -1;
        return rotOmega;
}

/**
 * Initialize the snow particles, ie. velocity and spiral movement radius
 */
extern "C"
void CUDA_Snow_InitParticles(int num_parts, float _gravity, float snowGrowth) {
        // Allocate main memory for rotation radiuses and omega values
        float  *rot_radius = (float *)malloc(sizeof(float) * 32);
        float  *rot_omega = (float *)malloc(sizeof(float) * 32);

        // Initialize random spiral radiuses and omega values
        for(int i = 0; i < 32; ++i) {
        	rot_radius[i] = myrandom()*2.0f;
            rot_omega[i] = calc_omega()*1.0f;
     	}

        // Copy attributes to GPU memory
        cudaMemcpyToSymbol(radiuses, rot_radius, 32*sizeof(float), 0, cudaMemcpyHostToDevice);
        cudaMemcpyToSymbol(omegas, rot_omega, 32*sizeof(float), 0, cudaMemcpyHostToDevice);
        cudaMemcpyToSymbol(snow_growth, &snowGrowth, sizeof(float), 0, cudaMemcpyHostToDevice);
        cudaMemcpyToSymbol(gravity, &_gravity, sizeof(float), 0, cudaMemcpyHostToDevice);

        // Free the memory for radiuses and omegas after they are copied to the GPU
        free(rot_radius);
        free(rot_omega);

        // Allocate memory for particle velocities
        CUDA_SAFE_CALL(cudaMalloc((void**)&part_vel, sizeof(float4) * num_parts));
        float4 *vel = (float4 *)malloc(sizeof(float4) * num_parts);

     	for(int i = 0; i < num_parts; ++i) {
        	float maxv = myrandom() + 1.0f;
                vel[i].x = myrandom();
                vel[i].y = -maxv;
                vel[i].z = myrandom();
                vel[i].w = 1.0f / (maxv * maxv);
	}

        // Copy the particle velocities to GPU memory and free the array in main memory
        CUDA_SAFE_CALL(cudaMemcpy(part_vel, vel, sizeof(float4) * num_parts, cudaMemcpyHostToDevice));
        free(vel);
}

/**
 * Free memory allocated to the particle velocities
 */
extern "C"
void CUDA_Snow_FreeParticles() {
        CUDA_SAFE_CALL(cudaFree(part_vel));
}

/**
 * Update the particle positions. Calculations are done by the GPU on a VBO
 */
extern "C"
void CUDA_Snow_MoveParticles(GLuint vbo, GLuint ovbo, int length, float deltaTime) {
        float4 *pos = 0;
        float4 *map = 0;
		
		// Map the bufferobject to into the address space of CUDA and return the base pointer of the mapping.
        CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&pos, vbo));
        CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&map, ovbo));
		//cudaGLMapBufferObject((void**)&pos, vbo);
		//cudaGLMapBufferObject((void**)&map, ovbo);

        cudaMemcpyToSymbol(delta_time, &deltaTime, sizeof(float), 0, cudaMemcpyHostToDevice);

		//setting cache preferance
		cudaFuncSetCacheConfig(part_update, cudaFuncCachePreferNone);

        // Call particle-update kernel on GPU
        dim3 block(PART_BLOCK, 1, 1);
        dim3 grid(length / block.x, 1, 1);
        part_update<<<grid, block>>>(pos, part_vel, map);

        // Call smooth ground kernel on GPU
        dim3 block2(SG_DIM, SG_DIM, 1);
        dim3 grid2(tdim/SG_DIM, tdim/SG_DIM, 1);
        smooth_ground<<<grid2, block2, SG_STRIDE*SG_STRIDE*sizeof(float4)>>>(map);

        CUDA_SAFE_CALL(cudaGLUnmapBufferObject(vbo));
        CUDA_SAFE_CALL(cudaGLUnmapBufferObject(ovbo));
		//cudaGLUnmapBufferObject(vbo);
		//cudaGLUnmapBufferObject(ovbo);
}

