#include "WindLBM.h"
#include "WindSystemLBM.cuh"

#include "CudaHelpers.h"

#include <iostream>

/**
 * Empty constructor, the init function MUST be run before any simulation is done
 */
WindLBM::WindLBM() {
	// Prevent deletion of memory that is not allocated
	has_allocated = false;
}

/**
 * Destructor, deletes memory both on CPU and GPU
 */
WindLBM::~WindLBM()
{
	std::cout << "Running LBM wind sim destructor." << std::endl;
	if(has_allocated) {
		std::cout << "LBM wind sim had allocated memory, cleaning up." << std::endl;
		CUDA_WindField_Free_LBM();

		// Free host memory for the particle distribution functions
		free(f0_h);
		free(f1_h);
		free(f2_h);
		free(f3_h);
		free(f4_h);
		free(f5_h);
		free(f6_h);
		free(f7_h);
		free(f8_h);
		free(f9_h);
		free(f10_h);
		free(f11_h);
		free(f12_h);
		free(f13_h);
		free(f14_h);
		free(f15_h);
		free(f16_h);
		free(f17_h);
		free(f18_h);

		glDeleteBuffers(1, &point_vbo);
	}
	std::cout << "LBM wind sim destructor done." << std::endl;
}

void WindLBM::Init(int sizeX, int sizeY, int sizeZ, unsigned int mapDim, float_type lbmOmega, float_type lbmViscosity, float4 velFactor, float_type initVel, float_type initRho, float4 boundaryVelFactor)
{
        std::cout << "WindLBM::Init()\n";

        // Lattice size and extra cells
        // ADDS TWO EXTRA CELLS IN EACH DIRECTION (same behaviour as SOR solver)
        extra = 0;
        size_x = sizeX + 2;
        size_y = sizeY + 2;
        size_z = sizeZ + 2;

	num_lbm_speeds = 4;
	lbm_speed_idx = 0;
	total_sim_time = 0;

	boundary_size = 1;

        velocity_factor = velFactor;
        boundary_velocity_factor = boundaryVelFactor;

	boundaries[0] = 0;
	boundaries[1] = 0;
	boundaries[2] = 0;

        // Defines the boundary speeds.
        // The windfield changes by the boundary-velocity varying between these values.
	lbm_speeds[0].x = 1.0f; lbm_speeds[0].y = 0.0f; lbm_speeds[0].z = 0.0f; lbm_speeds[0].w = 0.01;
	lbm_speeds[1].x = 2.0f; lbm_speeds[1].y = 0.0f; lbm_speeds[1].z = 1.0f; lbm_speeds[1].w = 0.05;
	lbm_speeds[2].x = 0.0f; lbm_speeds[3].y = 0.0f; lbm_speeds[3].z = 0.0f; lbm_speeds[3].w = 0.00;
	lbm_speeds[3].x = -1.0f; lbm_speeds[2].y = 0.0f; lbm_speeds[2].z = -1.0f; lbm_speeds[2].w = 0.05;

        for(int i=0;i<num_lbm_speeds;i++) {
                lbm_speeds[i].x *= boundary_velocity_factor.x;
                lbm_speeds[i].y *= boundary_velocity_factor.y;
                lbm_speeds[i].z *= boundary_velocity_factor.z;
                lbm_speeds[i].w *= boundary_velocity_factor.w;
        }

	// Relaxation parameter
	omega = lbmOmega;

	render_graphics = true;

        iter_count = 0;

	allocateMemoryCPU();
        std::cout << "LBM simulation allocated memory on CPU\n";

	initData();

        std::cout << "LBM simulation initialized data successfully\n";

	// Viscosity is the quantity that describes a fluid's resistance to flow
	viscosity = lbmViscosity;
        //viscosity = 43;
	
	allocateMemoryGPU(getSizeX(), getSizeY(), getSizeZ(), boundary_size, velocity_factor, f0_h, f1_h, f2_h, f3_h, f4_h, f5_h, f6_h, f7_h, f8_h, f9_h, f10_h, f11_h, f12_h, f13_h, f14_h, f15_h, f16_h, f17_h, f18_h);
        std::cout << "LBM simulation allocated memory on GPU\n";	

        // Apply an equal force to all lattice nodes in the X-direction
        // Method to move the liquid through porous rocks
	/*if(sizeof(float_type)==4)
	{ 
		force = 0.00001f; 
	} 
	else 
	{
		force = 0.00001;
	}*/
        force = 0.0;

        int fieldSize = (sizeX+2) * (sizeY+2) * (sizeZ+2);
        int fieldSizeBytes = fieldSize * sizeof(float4);

        point_count = fieldSize;

        glGenBuffers(1, &point_vbo);
        glBindBuffer(GL_ARRAY_BUFFER, point_vbo);
        glBufferData(GL_ARRAY_BUFFER, fieldSizeBytes*2, 0, GL_DYNAMIC_DRAW);
        glBindBuffer(GL_ARRAY_BUFFER, 0);

        CUDA_RegisterVBO(point_vbo);

	pressure_shader.load("pressure.vert", "pressure.frag");

        velocity_shader.load("velocity.vert", "velocity.frag");

        initial_rho = initRho;

        CUDA_WindField_Init_LBM(size_x, size_y, size_z, mapDim, initVel, initRho);

	printInitalState();

	has_allocated = true;
}

void WindLBM::SimulateWindField(int collideKernelType, float deltaTime, bool step, int frame, bool isRendering, float& mlups)
{
        #if PRINT_DEBUG
		printf("Sim windfield: Starting LBM-iteration\n");
	#endif
        static int count = 0;

	const float thresh = 50.0f;
        const float speed = 6.0f;
	
        total_sim_time += deltaTime;
        if(total_sim_time > thresh) {
        	++lbm_speed_idx;
            	total_sim_time = 0;
        }

	#if PRINT_DEBUG
		printf("TIME is: %f\n", total_sim_time);
	#endif

        // Interpolate between the boundary velocity vectors
	float3 a;
	a.x = lbm_speeds[lbm_speed_idx%num_lbm_speeds].x;
	a.y = lbm_speeds[lbm_speed_idx%num_lbm_speeds].y;
	a.z = lbm_speeds[lbm_speed_idx%num_lbm_speeds].z;

        float3 b;
        b.x = lbm_speeds[(lbm_speed_idx+1)%num_lbm_speeds].x;
        b.y = lbm_speeds[(lbm_speed_idx+1)%num_lbm_speeds].y;
        b.z = lbm_speeds[(lbm_speed_idx+1)%num_lbm_speeds].z;

	float vel_a = lbm_speeds[lbm_speed_idx%num_lbm_speeds].w;
        float vel_b = lbm_speeds[(lbm_speed_idx+1)%num_lbm_speeds].w;

        float3 boundaryFactor;
	//boundaryFactor.x = 0.06; boundaryFactor.y = 0.0; boundaryFactor.z = 0.0;
        boundaryFactor.x = (a.x*(thresh-total_sim_time)/thresh + b.x*total_sim_time/thresh) * speed;
        boundaryFactor.y = (a.y*(thresh-total_sim_time)/thresh + b.y*total_sim_time/thresh) * speed;
        boundaryFactor.z = (a.z*(thresh-total_sim_time)/thresh + b.z*total_sim_time/thresh) * speed;
	const float normalizeTo = (vel_a*(thresh-total_sim_time)/thresh + vel_b*total_sim_time/thresh) * speed;

        #if PRINT_DEBUG
		printf("Sim windfield: Boundary vector will be normalized to length: %f\n", normalizeTo);
                printf("Sim windfield: Boundary-vector before it is normalized: %f, %f, %f\n", boundaryFactor.x, boundaryFactor.y, boundaryFactor.z);
        #endif

	// Normalize boundary vector
	float boundaryVectorLength = sqrt(boundaryFactor.x * boundaryFactor.x + 
                                        boundaryFactor.y * boundaryFactor.y + 
                                        boundaryFactor.z * boundaryFactor.z);
	boundaryFactor.x = (boundaryFactor.x / boundaryVectorLength) * normalizeTo;
	boundaryFactor.y = (boundaryFactor.y / boundaryVectorLength) * normalizeTo;
	boundaryFactor.z = (boundaryFactor.z / boundaryVectorLength) * normalizeTo;
        
        #if PRINT_DEBUG
                printf("Sim windfield: Boundary-vector after it is normalized: %f, %f, %f\n", boundaryFactor.x, boundaryFactor.y, boundaryFactor.z);
        #endif
        
        if((step && count < frame) || !step) {
                if(step)
                        ++count;

                //struct timeval t1, t2;
                //struct timezone tz;

                // Runs the simulation C-function that calls the kernels
                // Rho on the boundary is set to the initial value
                CUDA_Simulate_WindField_LBM(collideKernelType, omega, deltaTime, step, frame, extra, boundary_size, boundaryFactor, initial_rho, force, isRendering, mlups);
               
                iter_count++;
        }
}

/**
 * Debug rendering of the obstacle field, with red boxes
 */
void WindLBM::RenderObstacles()
{
        CUDA_WindField_RenderObstacles_LBM(point_vbo, point_count);

        glColor4f(0.8f, 0.0f, 0.2f, 1.0f);
        glBindBuffer(GL_ARRAY_BUFFER, point_vbo);
        glVertexPointer(4, GL_FLOAT, 0, 0);
        glDrawArrays(GL_POINTS, 0, point_count);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
}

/**
 * Debug rendering of the windfield velocity vectors
 */
void WindLBM::RenderDebug(float limit)
{
        CUDA_WindField_RenderDebug_LBM(limit, point_vbo, point_count);

        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        velocity_shader.bind();
        glColor4f(0.0f, 0.0f, 1.0f, 1.0f);
        glBindBuffer(GL_ARRAY_BUFFER, point_vbo);
        glVertexPointer(4, GL_FLOAT, 0, 0);
        glDrawArrays(GL_LINES, 0, point_count*2);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        velocity_shader.unbind();
        glDisable(GL_BLEND);
}

/**
 * Debug rendering of the windfield pressure
 * Enables the shader, but this feature is not currently functional in this solver
 */
void WindLBM::RenderPressure()
{
        CUDA_WindField_RenderPressure_LBM(point_vbo, point_count);

        glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        pressure_shader.bind();
        glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
        glBindBuffer(GL_ARRAY_BUFFER, point_vbo);
        glVertexPointer(4, GL_FLOAT, 0, 0);
        glDrawArrays(GL_POINTS, 0, point_count);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        pressure_shader.unbind();
        glDisable(GL_BLEND);
        glDisable(GL_VERTEX_PROGRAM_POINT_SIZE);

}

/**
 * Initialize LBM data on the CPU
 * Legacy from the alternative CPU solver
 */
void WindLBM::initData()
{
	for (int z=0; z<getSizeZ(); z++) 
	{		
		for(int y=0; y<getSizeY(); y++)
		{
			for(int x=0; x<getSizeX(); x++)
			{
				int i = getIndex(x,y,z);
			
                                f0_h[i] = 0;
                                f1_h[i] = 0;
                                f2_h[i] = 0;
                                f3_h[i] = 0;
                                f4_h[i] = 0;
                                f5_h[i] = 0;
                                f6_h[i] = 0;
                                f7_h[i] = 0;
                                f8_h[i] = 0;
                                f9_h[i] = 0;
                                f10_h[i] = 0;
                                f11_h[i] = 0;
                                f12_h[i] = 0;
                                f13_h[i] = 0;
                                f14_h[i] = 0;
                                f15_h[i] = 0;
                                f16_h[i] = 0;
                                f17_h[i] = 0;
                                f18_h[i] = 0; 
			}		
		}
    }
}

/**
 * Get the lattice size in the X direction
 */
int WindLBM::getSizeX()
{
	return	size_x;
}

/**
 * Get the lattice size in the Y direction
 */
int WindLBM::getSizeY()
{
	return	size_y;
}

/**
 * Get the lattice size in the Z direction
 */
int WindLBM::getSizeZ()
{
	return	size_z;
}

/**
 * Allocate LBM memory on the CPU
 * Legacy from the alternative CPU solver
 */
void WindLBM::allocateMemoryCPU()
{
	// Allocate host memory for the particle distribution functions
	f0_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f1_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f2_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f3_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f4_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f5_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f6_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f7_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f8_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f9_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f10_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f11_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f12_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f13_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f14_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f15_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f16_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f17_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
	f18_h = (float_type*) malloc(getSizeX()*getSizeY()*getSizeZ()*sizeof(float_type));
}

/**
 * Toggle function for rendering
 * Used to disable rendering for debugging
 */
void WindLBM::doGraphics()
{
	render_graphics = !render_graphics;
}

/**
 * Returns whether rendering is enabled
 */
bool WindLBM::graphics()
{
	return render_graphics;
}

/**
 * Returns the current iteration count
 */
int WindLBM::getIteration()
{
	return iter_count;
}

/**
 * Prints information about the initial state
 */
void WindLBM::printInitalState()
{
	// Print simulation to console
        printf("Initial LBM parameters:\n");
	printf("Dimension: (%d,%d,%d) \n", size_x, size_y, size_z);
        printf("Million Lattice Locations: %f\n", (size_x * size_y * size_z) / (1000000.0));
	printf("Million Active Lattice Locations: %f\n", (countNumFluidNodes() / 1000000.0));
	printf("Omega: %lf \n", omega);
	printf("Kinematic Viscosity: %f\n", viscosity);
	printf("Force: %f \n", force);
	printf("Memory Usage: %f MB \n", (getSizeX()*getSizeY()*getSizeZ()*19*sizeof(float_type))/pow(1024.,2));
	
	if(sizeof(float_type)==4)
		printf("32-bit Precision \n");
	else
		printf("64-bit Precision \n");
}

/**
 * Returns the array index of a lattice node
 */
int WindLBM::getIndex(int x, int y, int z)
{
	return x + y*getSizeX() + z*getSizeX()*getSizeY();
}

