#include <stdlib.h>
#include <stdio.h>
//#include <math.h>
#include <string.h>
#include "defines.h"
#include "vmath.h"
#include "shader.h"
#include "opengl.h"
#include "utils.h"

#include <cutil.h>
#include <cuda_gl_interop.h>

#include <cutil_inline.h>
#include <cutil_gl_inline.h>

#include "CudaHelpers.h"

#include "WindSystem.cuh"
#include "WindSystemKernels.cu"
#include "ParticleSystem.cuh"



/**
 * Renders velocity lines for debugging
 */
void CUDA_WindField_RenderDebug(float limit, GLuint &point_vbo, int point_count) {
        Line *pos;

        // Map VBO
        cudaGLMapBufferObject((void**)&pos, point_vbo);

        // Call make velocity lines kernel on GPU
        dim3 block(dim.x, 1, 1);
        dim3 grid(point_count / dim.x, 1, 1);
        make_vel_lines<<<grid, block>>>(wind_vel, obstacle, pos, limit, idim);

        // Unmap VBO
        cudaGLUnmapBufferObject(point_vbo);
}

/**
 * Renders "boxes" to symbolize the obstacles (terrain) for debugging
 */
void CUDA_WindField_RenderObstacles(GLuint &point_vbo, int point_count) {
        float4 *pos;
        cudaGLMapBufferObject((void**)&pos, point_vbo);

        // Call kernel on GPU
        dim3 block(dim.x, 1, 1);
        dim3 grid(point_count / dim.x, 1, 1);
        make_obs_points<<<grid, block>>>(pos, obstacle, idim);
        cudaGLUnmapBufferObject(point_vbo);
}

/**
 * Renders the pressure field for debugging
 */
void CUDA_WindField_RenderPressure(GLuint &point_vbo, int point_count) {
        float4 *pos;
        cudaGLMapBufferObject((void**)&pos, point_vbo);

        // Call make pressure points kernel on GPU
        dim3 block(dim.x, 1, 1);
        dim3 grid(point_count / dim.x, 1, 1);
        make_pressure_points<<<grid, block>>>(pos, pressure, idim);
        cudaGLUnmapBufferObject(point_vbo);
}

void wind_to_texture(cudaPitchedPtr src) {
        cudaMemcpy3DParms parm = {0};
        parm.srcPtr = src;
        parm.dstArray = wind_vel_array;
        parm.extent = make_cudaExtent(dim.x, dim.z, dim.y);
        parm.kind = cudaMemcpyDeviceToDevice;
        CUDA_SAFE_CALL(cudaMemcpy3D(&parm));

        wind_vel_tex.filterMode = cudaFilterModeLinear;
        wind_vel_tex.addressMode[0] = cudaAddressModeClamp;
        wind_vel_tex.addressMode[1] = cudaAddressModeClamp;
        wind_vel_tex.addressMode[2] = cudaAddressModeClamp;
        cudaBindTextureToArray(wind_vel_tex, wind_vel_array);
}

void calc_table() {
        float __table[64];
        __table[63] = 0.0f;
        for(int i = 0; i < 63; ++i) {
                int div = 0;
                for(int j = 0; j < 6; ++j)
                        div += (i & (1<<j)) == 0;

                __table[i] = 1.0f / (float)div;
        }
        cudaMemcpyToSymbol(poisson_tab, __table, 64*sizeof(float), 0, cudaMemcpyHostToDevice);

        float temp[3] = {
                (float)dim.x / (float)SCENE_X,
                (float)dim.y / (float)SCENE_Y,
                (float)dim.z / (float)SCENE_Z
        };
        cudaMemcpyToSymbol(convert, temp, 3*sizeof(float), 0, cudaMemcpyHostToDevice);

        cudaMemcpyToSymbol(terrain_dim, &tdim, sizeof(int), 0, cudaMemcpyHostToDevice);
}

/**
 * Initialize the pressure field
 */
void wind_pressure_init() {
        int size = dim.x * dim.y * dim.z;
 
        pressure_extent = make_cudaExtent(dim.x*sizeof(float), dim.z, dim.y);
        CUDA_SAFE_CALL(cudaMalloc3D(&pressure, pressure_extent));
        
        float *p = new float[size];
        memset(p, 0, size*sizeof(float));

        // Copy pressure field to GPU
        {
                cudaMemcpy3DParms parm = {0};
                parm.srcPtr = make_cudaPitchedPtr(p, dim.x*sizeof(float), dim.x, dim.z);
                parm.dstPtr = pressure;
                parm.extent = pressure_extent;
                parm.kind = cudaMemcpyHostToDevice;
                CUDA_SAFE_CALL(cudaMemcpy3D(&parm));
        }

        delete [] p;
}

// Cleaned up simulation function (see original src for commented-out code)
void CUDA_WindField_Simulate(float deltaTime, bool step, int frame, int point_count, bool isRendering) {
	static int count = 0;
	static float time = 0;
	static int idx = 0;
	const float thresh = 15.0f;
	const float speed = 6.0f;
	time += deltaTime;
	if(time > thresh)
	{
		++idx;
	    time = 0;
	}
	float4 a = speeds[idx%NUM_SPEEDS];
	float4 b = speeds[(idx+1)%NUM_SPEEDS];
	float4 vec;
	vec.x = (a.x*(thresh-time)/thresh + b.x*time/thresh) * speed;
	vec.y = (a.y*(thresh-time)/thresh + b.y*time/thresh) * speed;
	vec.z = (a.z*(thresh-time)/thresh + b.z*time/thresh) * speed;
	vec.w = 1.0f;

	const float h = 1.0f;

	if((step && count < frame) || !step) {
		if(step)
			++count;

		dim3 block(dim.x, 1, 1);
		dim3 grid(point_count / dim.x, 1, 1);
		dim3 bgrid(dim.z, 1, 1);
		dim3 grid2(dim.z-2, 1, 1);

		dim3 block3(BLOCK_X, BLOCK_Z, 1);
		dim3 grid3(dim.x/BLOCK_X, dim.z/BLOCK_Z, 1);

		dim3 block33(BLOCK_X, BLOCK_Y, BLOCK_Z);
		dim3 grid33(dim.x/BLOCK_X, dim.y/BLOCK_Y, dim.z/BLOCK_Z);

//		int mask1 = 0;
//		int mask2 = 255;
		int mask2 = 127;

		wind_advect<<<grid, block, dim.x*sizeof(float4)>>>(wind_vel, pressure, obstacle, dim, deltaTime, vec);

		build_solution2<<<grid2, block, (dim.x+1)*sizeof(float4)>>>(wind_vel, solution, obstacle, dim, 0.5f / (1.0f * deltaTime));

		float relax[] = { 1.7f, 1.5f, 1.2f, 1.1f, 1.1f };
		for(int i = 0; i < 5; ++i) {
			solve_poisson2<<<grid2, block, 3*dim.x*sizeof(float)>>>(pressure, solution, obstacle, dim, relax[i], mask2);
			set_boundary2<<<bgrid, block, 0>>>(pressure, obstacle, dim, mask2);
		}

		wind_project2<<<grid2, block, (dim.x+1)*sizeof(float)>>>(wind_vel, pressure, dim, 2.0f*0.5f * deltaTime / h);
	
		if(isRendering)
			wind_to_texture(wind_vel);
	}
	cudaThreadSynchronize();
}

/**
 * Initialize wind simulation
 */
void CUDA_WindField_Init(int xdim, int ydim, int zdim, unsigned int mapdim) {
        dim.x = xdim + 2;
        dim.y = ydim + 2;
        dim.z = zdim + 2;

        idim.x = 1.0f / (float)dim.x;
        idim.y = 1.0f / (float)dim.y;
        idim.z = 1.0f / (float)dim.z;

	/*idim.x = 1.0f / (float)(SCENE_X + 2.0);
        idim.y = 1.0f / (float)(SCENE_Y + 2.0);
        idim.z = 1.0f / (float)(SCENE_Z + 2.0);*/
        
	tdim = mapdim;

        calc_table();

        int size = dim.x * dim.y * dim.z;
        int size_bytes = size * sizeof(float4);

        wind_pressure_init();
        CUDA_SAFE_CALL(cudaMalloc3D(&solution, pressure_extent));
        CUDA_SAFE_CALL(cudaMemset3D(solution, 0x00, pressure_extent));

        wind_extent = make_cudaExtent(dim.x*sizeof(float4), dim.z, dim.y);
        CUDA_SAFE_CALL(cudaMalloc3D(&wind_vel, wind_extent));
        CUDA_CheckError("cudaMalloc3D");

        obstacle_extent = make_cudaExtent(dim.x*sizeof(int), dim.z, dim.y);
        CUDA_SAFE_CALL(cudaMalloc3D(&obstacle, obstacle_extent));

        float4 *vel = (float4 *)malloc(size_bytes);
        memset(vel, 0, size_bytes);
        cudaMemcpy3DParms parm = {0};
        parm.srcPtr = make_cudaPitchedPtr(vel, dim.x*sizeof(float4), dim.x, dim.z);
        parm.dstPtr = wind_vel;
        parm.extent = wind_extent;
        parm.kind = cudaMemcpyHostToDevice;
        CUDA_SAFE_CALL(cudaMemcpy3D(&parm));
        CUDA_CheckError("cudaMemcpy3D");

        // Allocate the wind velocity field on the GPU
        // Bind the CUDA memory to a texture reference
        {
                cudaChannelFormatDesc desc = cudaCreateChannelDesc<float4>();
                cudaExtent extent = make_cudaExtent(dim.x, dim.z, dim.y);
                cudaMalloc3DArray(&wind_vel_array, &desc, extent);

                wind_to_texture(wind_vel);
                cudaThreadSynchronize();
        }

        free(vel);
}

void CUDA_WindField_Free() {
        CUDA_SAFE_CALL(cudaFree(wind_vel.ptr));
        CUDA_SAFE_CALL(cudaFree(pressure.ptr));
        CUDA_SAFE_CALL(cudaFree(solution.ptr));
        CUDA_SAFE_CALL(cudaFree(obstacle.ptr));
        CUDA_SAFE_CALL(cudaFreeArray(wind_vel_array));
}

/**
 * Updates obstacles for the wind field (primarily the ground)
 * @param ovbo buffer object ID
 * @param stage at which the update is in
 * @param y height threshold for checks
 */
void CUDA_WindField_UpdateObstacles(unsigned int ovbo, int stage, int y) {
        static float4 *map = 0;
        static int *obs = 0;

        // Create heightmap and obstacle arrays, if they dont exist
        if(!map)
                map = new float4[tdim*tdim];
        if(!obs)
                obs = new int[dim.x*dim.y*dim.z];
//        printf("updating obs... dim: %d %d %d, tdim: %d\n", dim.x, dim.y, dim.z, tdim);       

        // Choose update stage
        switch(stage) {
                // Copy the obstacle matrix from GPU memory to main memory
                case 0: {
                        float4 *dmap = 0;
                        CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&dmap, ovbo));
                        //cudaGLMapBufferObject((void**)&dmap, ovbo);
                        cudaMemcpy(map, dmap, sizeof(float4)*tdim*tdim, cudaMemcpyDeviceToHost);
                        CUDA_SAFE_CALL(cudaGLUnmapBufferObject(ovbo));
                        //cudaGLUnmapBufferObject(ovbo);
                        break;
                }

                // Reset the whole obstacle matrix to 0
                case 1: {
                        memset(obs, 0, dim.x*dim.y*dim.z*sizeof(int));
                        break;
                }

                // Set self bit of each cell to 1 if below terrain (plus snow),
                // 0 otherwise
                case 2: {
                        // Loop through the XZ-plane
                        for(int z = 1; z < dim.z-1; ++z) {
                                for(int x = 1; x < dim.x-1; ++x) {
                                        float h = SCENE_Y*(float)y/(float)dim.y;
                                        int sx = ((x-1) * tdim) / dim.x;
                                        int sz = ((z-1) * tdim) / dim.z;
                                        const float4 &v = map[J(sx,sz)];

                                        if(y) {
                                                // 1 if height is less, 0 if larger
                                                obs[I(x,y,z)] = h < (v.y + v.w);
                                        }
                                        else {
                                                obs[I(x,y,z)] = 127;
                                        }
                                        //obs[I(x,y,z)] = y ? (h < (v.y + v.w)) : 127;
                                }
                        }
                        break;
                }

                // Set boundaries (same as case 5)
                case 3: {
                        // Loop through the XY-plane
                        for(int x = 0; x < dim.x; ++x) {
                                for(int y = 0; y < dim.y; ++y) {
                                        obs[I(x,y,0)] = obs[I(x,y,1)];
                                        obs[I(x,y,dim.z-1)] = obs[I(x,y,dim.z-2)];

                                        obs[I(0,y,x)] = obs[I(1,y,x)];
                                        obs[I(dim.x-1,y,x)] = obs[I(dim.x-2,y,x)];
                                }
                        }
                        break;
                }

                // Store information about neighbouring cells
                case 4: {
                        y += 1;
                        for(int z = 1; z < dim.z-1; ++z) {
                                for(int x = 1; x < dim.x-1; ++x) {
                                        int v = obs[I(x,y,z)];

                                        v |= (obs[I(x-1,y,z)] & VOX_SELF) * VOX_LEFT;
                                        v |= (obs[I(x+1,y,z)] & VOX_SELF) * VOX_RIGHT;
                                        v |= (obs[I(x,y,z+1)] & VOX_SELF) * VOX_UP;
                                        v |= (obs[I(x,y,z-1)] & VOX_SELF) * VOX_DOWN;
                                        v |= (obs[I(x,y+1,z)] & VOX_SELF) * VOX_ABOVE;
                                        v |= (obs[I(x,y-1,z)] & VOX_SELF) * VOX_BELOW;

                                        v |= (obs[I(x-1,y-1,z-1)] & VOX_SELF) * (1 << 7);
                                        v |= (obs[I(x-1,y,z-1)] & VOX_SELF) * (1 << 8);
                                        v |= (obs[I(x-1,y+1,z-1)] & VOX_SELF) * (1 << 9);
                                        v |= (obs[I(x+1,y-1,z-1)] & VOX_SELF) * (1 << 10);
                                        v |= (obs[I(x+1,y,z-1)] & VOX_SELF) * (1 << 11);
                                        v |= (obs[I(x+1,y+1,z-1)] & VOX_SELF) * (1 << 12);

                                        v |= (obs[I(x-1,y-1,z+1)] & VOX_SELF) * (1 << 13);
                                        v |= (obs[I(x-1,y,z+1)] & VOX_SELF) * (1 << 14);
                                        v |= (obs[I(x-1,y+1,z+1)] & VOX_SELF) * (1 << 15);
                                        v |= (obs[I(x+1,y-1,z+1)] & VOX_SELF) * (1 << 16);
                                        v |= (obs[I(x+1,y,z+1)] & VOX_SELF) * (1 << 17);
                                        v |= (obs[I(x+1,y+1,z+1)] & VOX_SELF) * (1 << 18);

                                        v |= (obs[I(x-1,y-1,z)] & VOX_SELF) * (1 << 19);
                                        v |= (obs[I(x+1,y-1,z)] & VOX_SELF) * (1 << 20);
                                        v |= (obs[I(x,y-1,z-1)] & VOX_SELF) * (1 << 21);
                                        v |= (obs[I(x,y-1,z+1)] & VOX_SELF) * (1 << 22);
                                        v |= (obs[I(x-1,y+1,z)] & VOX_SELF) * (1 << 23);
                                        v |= (obs[I(x+1,y+1,z)] & VOX_SELF) * (1 << 24);
                                        v |= (obs[I(x,y+1,z-1)] & VOX_SELF) * (1 << 25);
                                        v |= (obs[I(x,y+1,z+1)] & VOX_SELF) * (1 << 26);

                                        obs[I(x,y,z)] = v;
                                }
                        }
                        break;
                }

                // Set boundaries (same as case 3)
                case 5: {
                        for(int x = 0; x < dim.x; ++x) {
                                for(int y = 0; y < dim.y; ++y) {
                                        obs[I(x,y,0)] = obs[I(x,y,1)];
                                        obs[I(x,y,dim.z-1)] = obs[I(x,y,dim.z-2)];

                                        obs[I(0,y,x)] = obs[I(1,y,x)];
                                        obs[I(dim.x-1,y,x)] = obs[I(dim.x-2,y,x)];
                                }
                        }
                        break;
                }

                // Copy the obstacle matrix from main memory to GPU memory
                case 6: {
                        cudaMemcpy3DParms parm = {0};
                        parm.srcPtr = make_cudaPitchedPtr(obs, dim.x*sizeof(int), dim.x, dim.z);
                        parm.dstPtr = obstacle;
                        parm.extent = obstacle_extent;
                        parm.kind = cudaMemcpyHostToDevice;
                        //CUDA_SAFE_CALL(cudaMemcpy3D(&parm));
                        cudaMemcpy3D(&parm);
                        break;
                }
        }
}
