#ifndef WINDSYSTEMKERNELS_CU
#define WINDSYSTEMKERNELS_CU

__constant__ float3 dirs[27] = {
    0,0,0,

    -1,0,0,
    1,0,0,
    0,0,1,
    0,0,-1,
    0,1,0,
    0,-1,0,

    -1,-1,-1,
    -1,0,-1,
    -1,1,-1,
    1,-1,-1,
    1,0,-1,
    1,1,-1,

    -1,-1,1,
    -1,0,1,
    -1,1,1,
    1,-1,1,
    1,0,1,
    1,1,1,

    -1,-1,0,
    1,-1,0,
    0,-1,-1,
    0,-1,1,
    -1,1,0,
    1,1,0,
    0,1,-1,
    0,1,1
};



__device__ float3 cross(float3 a, float3 b) {
    return make_float3(a.y*b.z - a.z*b.y, 
            a.z*b.x - a.x*b.z, 
            a.x*b.y - a.y*b.x);
}

__device__ float length(float3 n) {
    return sqrtf(n.x*n.x + n.y*n.y + n.z*n.z);
}

__device__ float3 normalize(float3 n) {
    float ilen = 1.0f / length(n);
    n.x *= ilen;
    n.y *= ilen;
    n.z *= ilen;
    return n;
}

__device__ float4 wind_vel_sample(float4 pos) {
    float x = pos.x + 1.0f;
    float z = pos.y + 1.0f;
    float y = pos.z + 1.0f;
    return tex3D(wind_vel_tex, x, y, z);
}

#define REF(T,p,x,y,z) (*((T *)((char *)(p).ptr + (p).pitch * (p).ysize * (y) + (p).pitch * (z)) + (x)))

#define INIT(p) \
    int x = threadIdx.x; \
int y = blockIdx.x / p.ysize; \
int z = blockIdx.x - (y * p.ysize)

__global__ void wind_advect(cudaPitchedPtr u, cudaPitchedPtr p, cudaPitchedPtr obs, const dim3 dim, const float dt, float4 boundary) {
    INIT(u); // calculate x, y, z coordinates

    float4 *v = (float4 *)shared + x;

    int mask = REF(int, obs, x, y, z);
    *v = REF(float4, u, x, y, z);

    volatile int zero = 0;

    if(x > zero && x < dim.x-1 && y > zero && y < dim.y-1 && z > zero && z < dim.z-1) {
        v->x = (float)x - dt * v->x - 0.5f;
        v->y = (float)y - dt * v->y - 0.5f;
        v->z = (float)z - dt * v->z - 0.5f;
        *v = wind_vel_sample(*v);

        if(mask & (VOX_SELF | VOX_LEFT | VOX_RIGHT)) v->x = zero;
        if(mask & (VOX_SELF | VOX_ABOVE | VOX_BELOW)) v->y = zero;
        if(mask & (VOX_SELF | VOX_UP | VOX_DOWN)) v->z = zero;

        if(y == dim.y-2)
            REF(float4, u, x, y+1, z) = *v;

        /*if(mask & VOX_SELF)
         *v = make_float4(0,0,0,1.0f);
         else if (mask != 0) {
         float3 n = make_float3(0, 0, 0);
#pragma unroll
for(int i = 1; i < 27; ++i) {
if(mask & (1 << i)) {
n.x += dirs[i].x;
n.y += dirs[i].y;
n.z += dirs[i].z;
}
}
n = normalize(n);
n.x = -n.x;
n.y = -n.y;
n.z = -n.z;

float3 v3 = make_float3(v->x, v->y, v->z);
n = cross(cross(n, v3), n);
v->x = n.x;
v->y = n.y;
v->z = n.z;
}*/
            }
else {
    //if(y != dim.y-1)
    *v = boundary;
}
REF(float4, u, x, y, z) = *v;
REF(float, p, x, y, z) = 0.0f;
}

/*
   __global__ void wind_advect(cudaPitchedPtr u, cudaPitchedPtr obs, const dim3 dim, const float dt, float4 boundary) {
   INIT(u); // calculate x, y, z coordinates

   int mask = REF(int, obs, x, y, z);
   float4 v = REF(float4, u, x, y, z);

   if(x > 0 && x < dim.x-1 && y > 0 && y < dim.y-1 && z > 0 && z < dim.z-1) {
   v.x = (float)x - dt * v.x - 0.5f;
   v.y = (float)y - dt * v.y - 0.5f;
   v.z = (float)z - dt * v.z - 0.5f;
   v = wind_vel_sample(v);

   if(mask & VOX_SELF)
   v = make_float4(0,0,0,1.0f);
   else if (mask != 0) {
   float3 n = make_float3(0, 0, 0);
#pragma unroll
for(int i = 1; i < 27; ++i) {
if(mask & (1 << i)) {
n.x += dirs[i].x;
n.y += dirs[i].y;
n.z += dirs[i].z;
}
}
n = normalize(n);
n.x = -n.x;
n.y = -n.y;
n.z = -n.z;

float3 v3 = make_float3(v.x, v.y, v.z);
n = cross(cross(n, v3), n);
v.x = n.x;
v.y = n.y;
v.z = n.z;
}
}
else {
v = boundary;
}
REF(float4, u, x, y, z) = v;
}
 */
template<typename T>
__global__ void set_boundary(cudaPitchedPtr p, dim3 dim, int which) {
    int x = threadIdx.x;
    int z = blockIdx.x;

    if(which & VOX_BELOW)
        REF(T, p, x, 0, z) = REF(T, p, x, 1, z);

    if(which & VOX_ABOVE)
        REF(T, p, x, dim.y-1, z) = REF(T, p, x, dim.y-2, z);

    if(blockIdx.x != 0 && blockIdx.x < dim.y-1) {
        if(which & VOX_UP)
            REF(T, p, x, z, 0) = REF(T, p, x, z, 1);

        if(which & VOX_DOWN)
            REF(T, p, x, z, dim.z-1) = REF(T, p, x, z, dim.z-2);
    }
}

__global__ void set_boundary2(cudaPitchedPtr p, cudaPitchedPtr obs, dim3 dim, int which) {
    INIT(p);

    int mask = REF(int, obs, x, y, z) & 127;

    if(x > 0 && x < dim.x-1 && y > 0 && y < dim.y-1 && z > 0 && z < dim.z-1) {
        if((mask & VOX_SELF) && mask != 127) {
            if(!(mask & VOX_LEFT))
                REF(float, p, x, y, z) = REF(float, p, x-1, y, z);
            else if(!(mask & VOX_RIGHT))
                REF(float, p, x, y, z) = REF(float, p, x+1, y, z);
            else if(!(mask & VOX_UP))
                REF(float, p, x, y, z) = REF(float, p, x, y, z-1);
            else if(!(mask & VOX_DOWN))
                REF(float, p, x, y, z) = REF(float, p, x, y, z+1);
            else if(!(mask & VOX_ABOVE))
                REF(float, p, x, y, z) = REF(float, p, x, y+1, z);
            else if(!(mask & VOX_BELOW))
                REF(float, p, x, y, z) = REF(float, p, x, y-1, z);
        }

        if(x == 1 && which & VOX_LEFT)
            REF(float, p, 0, y, z) = REF(float, p, 1, y, z);
        if(x == dim.x-2 && which & VOX_RIGHT)
            REF(float, p, dim.x-1, y, z) = REF(float, p, dim.x-2, y, z);

        if(z == 1 && which & VOX_UP)
            REF(float, p, x, y, 0) = REF(float, p, x, y, 1);
        if(z == dim.z-2 && which & VOX_DOWN)
            REF(float, p, x, y, dim.z-1) = REF(float, p, x, y, dim.z-2);

        if(y == 1 && which & VOX_BELOW)
            REF(float, p, x, 0, z) = REF(float, p, x, 1, z);
        if(y == dim.y-2 && which & VOX_ABOVE)
            REF(float, p, x, dim.y-1, z) = REF(float, p, x, dim.y-2, z);
    }
}


__global__ void set_boundary_vel(cudaPitchedPtr p, cudaPitchedPtr obs, dim3 dim, int which) {
    //float4 *ptr;
    INIT(p);

    int mask = REF(int, obs, x, y, z);

    if(x > 0 && x < dim.x-1 && y > 0 && y < dim.y-1 && z > 0 && z < dim.z-1) {
        if(mask & VOX_SELF)
            REF(float4, p, x, y, z) = make_float4(0,0,0,1.0f);
        /*else if (mask != 0) {
          float3 n = make_float3(0, 0, 0);
#pragma unroll
for(int i = 1; i < 27; ++i) {
if(mask & (1 << i)) {
n.x += dirs[i].x;
n.y += dirs[i].y;
n.z += dirs[i].z;
}
}
n = normalize(n);
n.x = -n.x;
n.y = -n.y;
n.z = -n.z;

ptr = &REF(float4, p, x, y, z);
float4 v = *ptr;
float3 v3 = make_float3(v.x, v.y, v.z);
n = cross(cross(n, v3), n);
v.x = n.x;
v.y = n.y;
v.z = n.z;
         *ptr = v;
         }*/

            /*if(x == 1 && which & VOX_LEFT)
              REF(float4, p, 0, y, z) = REF(float4, p, 1, y, z);
              else if(x == dim.x-2 && which & VOX_RIGHT)
              REF(float4, p, dim.x-1, y, z) = REF(float4, p, dim.x-2, y, z);*/
            }
}

__global__ void solve_poisson(cudaPitchedPtr p, cudaPitchedPtr b, cudaPitchedPtr obs, dim3 dim, float w, int which) {
    volatile int zero = 0;

    float *block = (float *)shared;
    float res = zero;

    INIT(p);

    int mask = REF(int, obs, x, y, z) & 127;
    block[x] = REF(float, p, x, y, z);
    __syncthreads();

    int div = zero;
#pragma unroll
    for(int i = 1; i <= 6; ++i)
        div += (mask & (1<<i)) == zero;

    if(div > zero) {
        /*if(((unsigned)(x-1) <= (unsigned)(dim.x+1)) *
          ((unsigned)(y-1) <= (unsigned)(dim.y+1)) *
          ((unsigned)(z-1) <= (unsigned)(dim.z+1))) {*/
        if(x > zero && x < dim.x-1 && y > zero && y < dim.y-1 && z > zero && z < dim.z-1) {
            if((mask & VOX_SELF) == zero) {
                if((mask & VOX_LEFT) == zero) res += block[x-1];
                if((mask & VOX_RIGHT) == zero) res += block[x+1];

                if((mask & VOX_UP) == zero) res += REF(float, p, x, y, z-1);
                if((mask & VOX_DOWN) == zero) res += REF(float, p, x, y, z+1);

                if((mask & VOX_BELOW) == zero) res += REF(float, p, x, y-1, z);
                if((mask & VOX_ABOVE) == zero) res += REF(float, p, x, y+1, z);

                res -= REF(float, b, x, y, z);

                res *= w / (float)div;
            }

            res += block[x] * (1.0f - w);

            /*if(x == 1 && which & VOX_LEFT)
              REF(float, p, 0, y, z) = res;
              else if(x == dim.x-2 && which & VOX_RIGHT)
              REF(float, p, dim.x-1, y, z) = res;*/

            REF(float, p, x, y, z) = res;
        }
    }
        else {
            REF(float, p, x, y, z) = zero;
        }
    }

    __constant__ float poisson_tab[64];

    __global__ void solve_poisson3(cudaPitchedPtr p, cudaPitchedPtr q, cudaPitchedPtr b, cudaPitchedPtr obs, dim3 dim, float w) {
        float *block = (float *)shared + (threadIdx.y+1)*BLOCK_W + threadIdx.x+1;
        float *prev = (float *)shared + BLOCK_SIZE + threadIdx.y*BLOCK_X + threadIdx.x;
        int *masks = (int *)(prev + BLOCK_X*BLOCK_Z) + threadIdx.y*BLOCK_X + threadIdx.x;

        obs.ptr = (int *)((char *)obs.ptr + (blockIdx.y * BLOCK_Z + threadIdx.y)*obs.pitch) + blockIdx.x * BLOCK_X + threadIdx.x;
        p.ptr = (float *)((char *)p.ptr + (blockIdx.y * BLOCK_Z + threadIdx.y)*p.pitch) + blockIdx.x * BLOCK_X + threadIdx.x;
        b.ptr = (float *)((char *)b.ptr + (blockIdx.y * BLOCK_Z + threadIdx.y)*b.pitch) + blockIdx.x * BLOCK_X + threadIdx.x;

        LOAD(0)
            prev[0] = block[0];
        LOAD(1)

            for(int y = 1; y < dim.y-1; ++y) {
                obs.ptr = (char *)obs.ptr + obs.ysize*obs.pitch;
                p.ptr = (char *)p.ptr + p.ysize*p.pitch;
                b.ptr = (char *)b.ptr + b.ysize*b.pitch;

                __syncthreads();

                masks[0] = *(int *)obs.ptr & 127;

                float res = 0.0f;

                if(~masks[0] & 126) {
                    if((masks[0] & VOX_LEFT) == 0) res += block[-1];
                    if((masks[0] & VOX_RIGHT) == 0) res += block[1];
                    if((masks[0] & VOX_UP) == 0) res += block[-BLOCK_W];
                    if((masks[0] & VOX_DOWN) == 0) res += block[BLOCK_W];
                }

                float curr = block[0];
                __syncthreads();
                LOAD(1)

                    if(~masks[0] & 126) {
                        if((masks[0] & VOX_BELOW) == 0) res += prev[0];
                        if((masks[0] & VOX_ABOVE) == 0) res += block[0];

                        res -= *(float *)b.ptr;
                        res *= w * poisson_tab[masks[0]>>1];
                        res += curr * (1.0f - w);
                    }

                *(float *)p.ptr = res;
                prev[0] = curr;
            }
    }

    __global__ void solve_poisson2(cudaPitchedPtr p, cudaPitchedPtr b, cudaPitchedPtr obs, dim3 dim, float w, int which) {
        float *block = (float *)shared;
        float *prev = block + dim.x;
        int *masks = (int *)prev + dim.x;
        float curr;

        volatile int zero = 0;
        volatile int one = 1;

        int x = threadIdx.x;
        int z = blockIdx.x+one;

        prev[x] = REF(float, p, x, dim.y-one, z);
        block[x] = REF(float, p, x, dim.y-2, z);

        //#pragma unroll
        for(int y = dim.y-2; y > zero; --y) {
            __syncthreads();

            masks[x] = REF(int, obs, x, y, z) & 127;

            if((~masks[x] & 126) && (masks[x] & VOX_SELF) == zero) {
                float res;

                if((unsigned)(x-one) <= (unsigned)(dim.x+one)) {
                    res = zero;
                    res += block[x-one] * (float)((masks[x] & VOX_LEFT) == zero);
                    res += block[x+one] * (float)((masks[x] & VOX_RIGHT) == zero);

                    res += REF(float, p, x, y, z-one) * (float)((masks[x] & VOX_UP) == zero);
                    res += REF(float, p, x, y, z+one) * (float)((masks[x] & VOX_DOWN) == zero);

                    curr = block[x];
                }

                __syncthreads();
                block[x] = REF(float, p, x, y-one, z);

                if((unsigned)(x-one) <= (unsigned)(dim.x+one)) {
                    res += prev[x] * (float)((masks[x] & VOX_BELOW) == zero);
                    res += block[x] * (float)((masks[x] & VOX_ABOVE) == zero);
                    res -= REF(float, b, x, y, z);
                    res *= poisson_tab[masks[x]>>one];
                    res *= w;
                    res += curr * (1.0f - w);
                    prev[x] = curr;

                    //				if(x == 1 && which & VOX_LEFT)
                    //					REF(float, p, 0, y, z) = res;
                    //				else if(x == dim.x-2 && which & VOX_RIGHT)
                    //					REF(float, p, dim.x-1, y, z) = res;

                    REF(float, p, x, y, z) = res;
                }
            }
            else {
                REF(float, p, x, y, z) = 0.0f;
            }
        }
    }

    __global__ void solve_poisson33(cudaPitchedPtr p, cudaPitchedPtr q, cudaPitchedPtr b, cudaPitchedPtr obs, dim3 dim, float w) {
        float *block = (float *)shared;
        float res;
        int x, y, z, mask, div;

        x = blockIdx.x * BLOCK_X + threadIdx.x;
        y = blockIdx.y * BLOCK_Y + threadIdx.y;
        z = blockIdx.z * BLOCK_Z + threadIdx.z;

        LOAD3(0)
            __syncthreads();

        mask = REF(int, obs, x, y, z) & 127;
        div = 0;
        div += (mask & VOX_LEFT) == 0;
        div += (mask & VOX_RIGHT) == 0;
        div += (mask & VOX_UP) == 0;
        div += (mask & VOX_DOWN) == 0;
        div += (mask & VOX_BELOW) == 0;
        div += (mask & VOX_ABOVE) == 0;
        div *= (mask & VOX_SELF) == 0;

        res = 0.0f;

        if(div > 0) {
            if((mask & VOX_LEFT) == 0) res += STENCIL3(-1,0,0);
            if((mask & VOX_RIGHT) == 0) res += STENCIL3(1,0,0);
            if((mask & VOX_UP) == 0) res += STENCIL3(0,0,-1);
            if((mask & VOX_DOWN) == 0) res += STENCIL3(0,0,1);
            if((mask & VOX_BELOW) == 0) res += STENCIL3(0,-1,0);
            if((mask & VOX_ABOVE) == 0) res += STENCIL3(0,1,0);

            res -= REF(float, b, x, y, z);
            res *= w / (float)div;
            res += STENCIL3(0,0,0) * (1.0f - w);
        }

        REF(float, p, x, y, z) = res;
    }


    __global__ void solve_poisson22(cudaPitchedPtr p, cudaPitchedPtr b, cudaPitchedPtr obs, dim3 dim, float w, int which) {
        float *block = (float *)shared;
        float *prev = block + dim.x;
        float res;
        float curr;
        int mask, div;

        int x = threadIdx.x;
        int z = blockIdx.x+1;

        prev[x] = REF(float, p, x, 0, z);
        block[x] = REF(float, p, x, 1, z);

        //#pragma unroll
        for(int y = 1; y < dim.y-1; ++y) {
            __syncthreads();

            mask = REF(int, obs, x, y, z) & 127;
            div = 0;
#pragma unroll
            for(int i = 1; i <= 6; ++i)
                div += (mask & (1<<i)) == 0;

            if(div > 0) {
                if(x > 0 && x < dim.x-1) {
                    res = 0;
                    res += block[x-1] * (float)((mask & VOX_LEFT) == 0);
                    res += block[x+1] * (float)((mask & VOX_RIGHT) == 0);

                    res += REF(float, p, x, y, z-1) * (float)((mask & VOX_UP) == 0);
                    res += REF(float, p, x, y, z+1) * (float)((mask & VOX_DOWN) == 0);

                    curr = block[x];
                }

                __syncthreads();
                block[x] = REF(float, p, x, y+1, z);

                if(x > 0 && x < dim.x-1) {
                    res += prev[x] * (float)((mask & VOX_BELOW) == 0);
                    res += block[x] * (float)((mask & VOX_ABOVE) == 0);
                    res -= REF(float, b, x, y, z);
                    res *= (float)((mask & VOX_SELF) == 0) * w / (float)div;
                    res += curr * (1.0f - w);
                    prev[x] = curr;

                    /*				if(x == 1 && which & VOX_LEFT)
                                    REF(float, p, 0, y, z) = res;
                                    else if(x == dim.x-2 && which & VOX_RIGHT)
                                    REF(float, p, dim.x-1, y, z) = res;*/

                    REF(float, p, x, y, z) = res;
                }
            }
            else {
                REF(float, p, x, y, z) = 0.0f;
            }
        }
    }

    __global__ void build_solution2(cudaPitchedPtr u, cudaPitchedPtr b, cudaPitchedPtr obs, dim3 dim, float factor) {
        float4 *block = (float4 *)shared + threadIdx.x;

        //volatile int _0 = 0;
        volatile int _1 = 1;

        //int active = (unsigned)(threadIdx.x-_1) <= (unsigned)(dim.x+_1);
        int active = (unsigned)(threadIdx.x-_1) <= (unsigned)(dim.x+_1) && threadIdx.x < dim.x-1;
        //int active = (unsigned)(threadIdx.x-_1) <= (unsigned)(dim.x+_1);
        obs.ptr = (int *)((char *)obs.ptr + (blockIdx.x+_1)*obs.pitch) + threadIdx.x;
        b.ptr = (float *)((char *)b.ptr + (blockIdx.x+_1)*b.pitch) + threadIdx.x;
        u.ptr = (float4 *)((char *)u.ptr + (blockIdx.x+_1)*u.pitch) + threadIdx.x;

        block[0] = *(float4 *)u.ptr;
        float last = block[0].y;
        u.ptr = (char *)u.ptr + u.pitch*u.ysize;
        block[0] = *(float4 *)u.ptr;

        for(int y = _1; y < dim.y-_1; ++y) {
            float res = -last;
            last = block[0].y;

            __syncthreads();
            if(active) {
                res += block[_1].x;
                res -= block[-1].x;
            }
            __syncthreads();

            block[0] = *(float4 *)((char *)u.ptr + u.pitch);
            res += block[0].z;
            block[0] = *(float4 *)((char *)u.ptr - u.pitch);
            res -= block[0].z;

            u.ptr = (char *)u.ptr + u.pitch*u.ysize;
            block[0] = *(float4 *)u.ptr;

            res += block[0].y;
            res *= factor;

            obs.ptr = (char *)obs.ptr + obs.pitch*obs.ysize;
            res *= (float)((*(int *)obs.ptr & VOX_SELF) == 0);

            b.ptr = (char *)b.ptr + b.pitch*b.ysize;

            if(active) {
                *(float *)b.ptr = res;
            }
        }
    }

    __global__ void build_solution22(cudaPitchedPtr u, cudaPitchedPtr b, cudaPitchedPtr p, cudaPitchedPtr obs, dim3 dim, float factor, int which) {
        // BEWARE the deleted commented-out code
        float *block = (float *)shared;
        float res;
        float last;

        int x = threadIdx.x;
        int z = blockIdx.x+1;

        float4 *v = (float4 *)(block + dim.x) + x;
        *v = REF(float4, u, x, 0, z);
        block[x] = v->x;
        last = v->y;
        *v = REF(float4, u, x, 1, z);
        block[x] = v->x;

        //#pragma unroll
        for(int y = 1; y < dim.y-1; ++y) {
            __syncthreads();

            if((unsigned)(x-1) <= (unsigned)(dim.x+1)) {
                res = 0;
                res += (block[x+1] - block[x-1]);

                *v = REF(float4, u, x, y, z+1);
                res += v->z;
                *v = REF(float4, u, x, y, z-1);
                res -= v->z;
            }

            __syncthreads();
            *v = REF(float4, u, x, y+1, z);
            block[x] = v->x;

            if((unsigned)(x-1) <= (unsigned)(dim.x+1)) {
                res += (v->y - last);
                res *= factor;
                res *= (float)((REF(int, obs, x, y, z) & VOX_SELF) == 0);
                last = v->y;

                REF(float, b, x, y, z) = res;
            }

            REF(float, p, x, y, z) = 0.0f;
        }
    }

    __global__ void build_solution3(cudaPitchedPtr u, cudaPitchedPtr b, cudaPitchedPtr p, cudaPitchedPtr obs, dim3 dim, float factor, int which) {
        float4 *block = (float4 *)shared;
        float4 *vec = block+dim.x;
        float res;
        float last, curr;

        int x = threadIdx.x;
        int z = blockIdx.x+1;

        vec[x] = REF(float4, u, x, 0, z);
        last =  vec[x].y;
        block[x] = REF(float4, u, x, 1, z);

        //#pragma unroll
        for(int y = 1; y < dim.y-1; ++y) {
            __syncthreads();

            if((unsigned)(x-1) <= (unsigned)(dim.x+1)) {
                res = 0;
                res += (block[x+1].x - block[x-1].x);

                vec[x] = REF(float4, u, x, y, z+1);
                res += vec[x].z;
                vec[x] = REF(float4, u, x, y, z-1);
                res -= vec[x].z;
                //res += (REF(float4, u, x, y, z+1).z - REF(float4, u, x, y, z-1).z);
                curr = block[x].y;
            }

            __syncthreads();
            block[x] = REF(float4, u, x, y+1, z);

            if((unsigned)(x-1) <= (unsigned)(dim.x+1)) {
                res += (block[x].y - last);
                res *= factor;
                res *= (float)((REF(int, obs, x, y, z) & VOX_SELF) == 0);
                last = curr;

                if(x == 1 && which & VOX_LEFT)
                    REF(float, b, 0, y, z) = res;
                if(x == dim.x-2 && which & VOX_RIGHT)
                    REF(float, b, dim.x-1, y, z) = res;

                REF(float, b, x, y, z) = res;
            }

            REF(float, p, x, y, z) = 0.0f;
        }
    }

    __global__ void build_solution(cudaPitchedPtr u, cudaPitchedPtr b, cudaPitchedPtr p, cudaPitchedPtr obs, dim3 dim, float h, float dt, int which) {
        float4 *block = (float4 *)shared;
        float h_inv = 0.5f / (h * dt);

        INIT(p);

        int mask = REF(int, obs, x, y, z) & 127;
        block[x] = REF(float4, u, x, y, z);
        __syncthreads();
        if(x > 0 && x < dim.x-1 && y > 0 && y < dim.y-1 && z > 0 && z < dim.z-1) {
            float res = 0;

            if((mask & VOX_SELF) == 0) {
                res += (block[x+1].x - block[x-1].x);
                res += (REF(float4, u, x, y, z+1).z - REF(float4, u, x, y, z-1).z);
                res += (REF(float4, u, x, y+1, z).y - REF(float4, u, x, y-1, z).y);
                res *= h_inv;
            }

            /*if(x == 1 && which & VOX_LEFT)
              REF(float, b, 0, y, z) = res;
              if(x == dim.x-2 && which & VOX_RIGHT)
              REF(float, b, dim.x-1, y, z) = res;*/

            REF(float, b, x, y, z) = res;
        }

        REF(float, p, x, y, z) = 0.0f;
    }

    __global__ void build_solution_ns(cudaPitchedPtr u, cudaPitchedPtr c, cudaPitchedPtr b, cudaPitchedPtr p, cudaPitchedPtr obs, dim3 dim, float h, float dt, int which) {
        float4 *block = (float4 *)shared;
        float res = 0;
        float h_inv = 0.5f / (h * dt);
        float h2_inv = 0.05f / (h * h);

        INIT(p);

        int mask = REF(int, obs, x, y, z) & 127;
        block[x] = REF(float4, u, x, y, z);
        __syncthreads();
        if(x > 0 && x < dim.x-1 && y > 0 && y < dim.y-1 && z > 0 && z < dim.z-1) {
            float4 left = block[x-1];
            float4 right = block[x+1];
            float4 down = REF(float4, u, x, y, z+1);
            float4 up = REF(float4, u, x, y, z-1);
            float4 below = REF(float4, u, x, y-1, z);
            float4 above = REF(float4, u, x, y+1, z);
            float4 v = block[x];

            float3 curl;
            curl.x = 0.5f*(above.z - below.z) - 0.5f*(down.y - up.y);
            curl.y = 0.5f*(down.x - up.x) - 0.5f*(right.z - left.z);
            curl.z = 0.5f*(right.y - left.y) - 0.5f*(above.x - below.x);

            float3 n;
            n.x = 0.5f*(REF(float, c, x+1, y, z) - REF(float, c, x-1, y, z));
            n.y = 0.5f*(REF(float, c, x, y+1, z) - REF(float, c, x, y-1, z));
            n.z = 0.5f*(REF(float, c, x, y, z+1) - REF(float, c, x, y, z-1));
            float len = length(n);
            if(len > 0.00001f) len = 1.0f / len;
            else len = 1.0f;
            n.x *= len;
            n.y *= len;
            n.z *= len;



            v.x += h2_inv*dt*(-6.0f*v.x + (right.x + left.x) + (up.x + down.x) + (above.x + below.x));
            v.y += h2_inv*dt*(-6.0f*v.y + (right.y + left.y) + (up.y + down.y) + (above.y + below.y));
            v.z += h2_inv*dt*(-6.0f*v.z + (right.z + left.z) + (up.z + down.z) + (above.z + below.z));

            n = cross(curl, n);
            float e = 0.10f;
            v.x -= e*n.x;
            v.y -= e*n.y;
            v.z -= e*n.z;

            REF(float4, u, x, y, z) = v;

            res += (right.x - left.x);
            res += (down.z - up.z);
            res += (above.y - below.y);
            res *= h_inv;
            res *= (float)((mask & VOX_SELF) == 0);

            if(x == 1 && which & VOX_LEFT)
                REF(float, b, 0, y, z) = res;
            if(x == dim.x-2 && which & VOX_RIGHT)
                REF(float, b, dim.x-1, y, z) = res;

            REF(float, b, x, y, z) = res;
        }

        REF(float, p, x, y, z) = 0.0f;
    }

    __global__ void build_curl(cudaPitchedPtr u, cudaPitchedPtr c, dim3 dim, float h) {
        float4 *block = (float4 *)shared;
        //float res = 0;
        //float h_inv = 0.5f / (h * 1.0f);
        //float h2_inv = 1.0f / (h * h);

        INIT(u);

        block[x] = REF(float4, u, x, y, z);
        __syncthreads();
        if(x > 0 && x < dim.x-1 && y > 0 && y < dim.y-1 && z > 0 && z < dim.z-1) {
            float4 left = block[x-1];
            float4 right = block[x+1];
            float4 down = REF(float4, u, x, y, z+1);
            float4 up = REF(float4, u, x, y, z-1);
            float4 below = REF(float4, u, x, y-1, z);
            float4 above = REF(float4, u, x, y+1, z);
            //float4 v = block[x];

            float3 curl;
            curl.x = 0.5f*(above.z - below.z) - 0.5f*(down.y - up.y);
            curl.y = 0.5f*(down.x - up.x) - 0.5f*(right.z - left.z);
            curl.z = 0.5f*(right.y - left.y) - 0.5f*(above.x - below.x);

            float mag = length(curl);
            REF(float, c, x, y, z) = mag;


            if(x == 1)
                REF(float, c, 0, y, z) = mag;
            if(x == dim.x-2)
                REF(float, c, dim.x-1, y, z) = mag;

        }
    }


    __global__ void wind_project2(cudaPitchedPtr u, cudaPitchedPtr p, dim3 dim, float factor) {
        float *block = (float *)shared + threadIdx.x;

        volatile int _1 = 1;

        int active = (unsigned)(threadIdx.x-_1) <= (unsigned)(dim.x+_1) && threadIdx.x < dim.x-1;
//        int active = (unsigned)(threadIdx.x-_1) <= (unsigned)(dim.x+_1);
        u.ptr = (float4 *)((char *)u.ptr + (blockIdx.x+_1)*u.pitch) + threadIdx.x;
        p.ptr = (float *)((char *)p.ptr + (blockIdx.x+_1)*p.pitch) + threadIdx.x;

        float last = *(float *)p.ptr;
        p.ptr = (char *)p.ptr + p.pitch*p.ysize;
        block[0] = *(float *)p.ptr;

        for(int y = _1; y < dim.y-_1; ++y) {
            u.ptr = (char *)u.ptr + u.pitch*u.ysize;
            float4 v = *(float4 *)u.ptr;
            v.z -= factor * *(float *)((char *)p.ptr + p.pitch);
            v.z += factor * *(float *)((char *)p.ptr - p.pitch);

            __syncthreads();
            if(active) {
                v.x -= factor * block[_1];
                v.x += factor * block[-1];
            }
            __syncthreads();

            v.y += factor * last;
            last = block[0];
            p.ptr = (char *)p.ptr + p.pitch*p.ysize;
            block[0] = *(float *)p.ptr;
            v.y -= factor * block[0];

            if(active) {
                *(float4 *)u.ptr = v;
            }
        }
    }

    __global__ void wind_project(cudaPitchedPtr u, cudaPitchedPtr p, dim3 dim, float h, float dt) {
        float *block = (float *)shared;
        float factor = 0.5f * dt / h;

        INIT(u);

        block[x] = REF(float, p, x, y, z);
        __syncthreads();
        float4 v = REF(float4, u, x, y, z);

        if(x > 0 && x < dim.x-1 && y > 0 && y < dim.y-1 && z > 0 && z < dim.z-1) {
            v.x -= factor * (block[x+1] - block[x-1]);
            v.z -= factor * (REF(float, p, x, y, z+1) - REF(float, p, x, y, z-1));
            v.y -= factor * (REF(float, p, x, y+1, z) - REF(float, p, x, y-1, z));
        }

        REF(float4, u, x, y, z) = v;
    }

    struct Line {
        float4 p0, p1;
    };

    __global__ void make_vel_lines(cudaPitchedPtr vel, cudaPitchedPtr obs, Line *out, float limit, float3 idim) {
        int tid = __mul24(blockIdx.x, blockDim.x) + threadIdx.x;

        INIT(obs);

        volatile float4 v = REF(float4, vel, x, y, z);
        //int mask = REF(int, obs, x, y, z) & 127;

        float speed = sqrtf(v.x*v.x + v.y*v.y + v.z*v.z);
        float temp = speed;

        //if(temp < limit) {
        //	temp = 0.0f;
        //}
        //else {
        if(temp < 1.0f)
            temp = 1.0f;
        else
            temp = 1.0f / temp;
        //}

        //if(!mask) temp = 0.0f;

        Line l;

        l.p0.x = (float)x;
        l.p0.y = (float)y;
        l.p0.z = (float)z;
        l.p0.w = 1.0f;

        l.p1.x = l.p0.x + v.x*temp;
        l.p1.y = l.p0.y + v.y*temp;
        l.p1.z = l.p0.z + v.z*temp;
        l.p1.w = 1.0f;

        l.p0.x *= idim.x;
        l.p1.x *= idim.x;
        l.p0.y *= idim.y;
        l.p1.y *= idim.y;
        l.p0.z *= idim.z;
        l.p1.z *= idim.z;

        //volatile float4 vv = wind_vel_sample(l.p1);
        float4 vv = wind_vel_sample(l.p1);
        float speed2 = sqrtf(vv.x*vv.x + vv.y*vv.y + vv.z*vv.z);
        l.p0.w = l.p1.w = speed2 - speed;

        out[tid] = l;
    }

    __global__ void make_obs_points(float4 *out, cudaPitchedPtr obs, float3 idim) {
        int tid = __mul24(blockIdx.x, blockDim.x) + threadIdx.x;

        INIT(obs);

        int mask = REF(int, obs, x, y, z);

        float4 v;
        v.x = 0.0f;
        v.y = 0.0f;
        v.z = 0.0f;
        v.w = 1.0f;

        //mask &= VOX_SELF;

        // Store the position in the texture if the self-bit is set
        // Normalize coordinates to 0.0-1.0 for storing in the texture
        if(mask) {
            v.x = (float)x*idim.x;
            v.y = (float)y*idim.y;
            v.z = (float)z*idim.z;
        }

        out[tid] = v;
    }

    __global__ void make_pressure_points(float4 *out, cudaPitchedPtr p, float3 idim) {
        int tid = __mul24(blockIdx.x, blockDim.x) + threadIdx.x;

        INIT(p);

        float val = REF(float, p, x, y, z);

        out[tid] = make_float4((float)x*idim.x, (float)y*idim.y, (float)z*idim.z, val);
    }

#endif
