#include <stdlib.h>
#include <stdio.h>

#include <string.h>
#include "defines.h"
#include "vmath.h"
#include "shader.h"
#include "opengl.h"
#include "utils.h"
#include <cutil.h>
#include <cuda_gl_interop.h>
#include <cutil_inline.h>
#include <cutil_gl_inline.h>
#include "CudaHelpers.h"

#include "LBM_types.h"

#include "WindSystemLBM.cuh"
#include "WindSystemKernelsLBM.cu"
#include "ParticleSystem.cuh"

/**
 * Renders velocity lines for debugging
 */
void CUDA_WindField_RenderDebug_LBM(float limit, GLuint &point_vbo, int point_count) {
        Line *pos;

        // Map VBO
        cudaGLMapBufferObject((void**)&pos, point_vbo);

        // Call make velocity lines kernel on GPU
        dim3 block(dim.x, 1, 1);
        dim3 grid(point_count / dim.x, 1, 1);
        make_vel_lines_lbm<<<grid, block>>>(wind_vel, obstacle, pos, limit, idim);

        // Unmap VBO
        cudaGLUnmapBufferObject(point_vbo);
}

/**
 * Renders "boxes" to symbolize the obstacles (terrain) for debugging
 */
void CUDA_WindField_RenderObstacles_LBM(GLuint &point_vbo, int point_count) {
        float4 *pos;
        cudaGLMapBufferObject((void**)&pos, point_vbo);

        // Call kernel on GPU
        dim3 block(dim.x, 1, 1);
        dim3 grid(point_count / dim.x, 1, 1);
        make_obs_points<<<grid, block>>>(pos, obstacle, idim);
        cudaGLUnmapBufferObject(point_vbo);
}

/**
 * Renders the pressure field for debugging
 */
void CUDA_WindField_RenderPressure_LBM(GLuint &point_vbo, int point_count) {
        dim3 blockSize(dim.x);
        dim3 nBlocks(dim.y, dim.z);
	float4 *pos;
        cudaGLMapBufferObject((void**)&pos, point_vbo);

        // Call make pressure points kernel on GPU
	createPressureField_Kernel<<<nBlocks, blockSize>>>(pos, obstacle, pdf, dim, idim);
        cudaGLUnmapBufferObject(point_vbo);
}

/**
 * Initialize wind simulation
 */
void CUDA_WindField_Init_LBM(int xdim, int ydim, int zdim, unsigned int mapdim, float_type initVel, float_type initRho) {
	// This is set for use in snow sim code
	// use dimX/Y/Z for LBM size (set in allocateMemoryGPU)
	dim.x = xdim; 
	dim.y = ydim; 
	dim.z = zdim;

        tdim = mapdim;

	idim.x = 1.0f / (float)dim.x;
        idim.y = 1.0f / (float)dim.y;
        idim.z = 1.0f / (float)dim.z;

	// From calc_table, some constants
        float temp[3] = {
                (float)dim.x / (float)SCENE_X,
                (float)dim.y / (float)SCENE_Y,
                (float)dim.z / (float)SCENE_Z
        };
        cudaMemcpyToSymbol(convert, temp, 3*sizeof(float), 0, cudaMemcpyHostToDevice);
        cudaMemcpyToSymbol(terrain_dim, &tdim, sizeof(int), 0, cudaMemcpyHostToDevice);

        int size = dim.x * dim.y * dim.z;
        int size_bytes = size * sizeof(float4); 

        // Allocate memory for the windfield (velocities)
        wind_extent = make_cudaExtent(dimX*sizeof(float4), dimZ, dimY);
		//cudaPitchedPtr test;

        CUDA_SAFE_CALL(cudaMalloc3D(&wind_vel, wind_extent));
		//CUDA_SAFE_CALL(cudaMalloc3D(&test, wind_extent));
        CUDA_CheckError("cudaMalloc3D (CUDA_WindField_Init_LBM)");

        // Allocate memory the obstacles
	#if PRINT_DEBUG
		printf("Allocating obstacle, normal dim: %dx%dx%d, lbm dim: %dx%dx%d\n", dim.x, dim.y, dim.z, dimX, dimY, dimZ);
	#endif
        obstacle_extent = make_cudaExtent(dim.x*sizeof(int), dim.z, dim.y);
        CUDA_SAFE_CALL(cudaMalloc3D(&obstacle, obstacle_extent));

        // Initialize the windfield to zero
        // Not strictly necessary with LBM, but can be useful while testing
        float4 *vel = (float4 *)malloc(dimX * dimY * dimZ * sizeof(float4));
        memset(vel, 0, dimX * dimY * dimZ * sizeof(float4));
        cudaMemcpy3DParms parm = {0};
        parm.srcPtr = make_cudaPitchedPtr(vel, dimX*sizeof(float4), dimX, dimZ);
        parm.dstPtr = wind_vel;
        parm.extent = wind_extent;
        parm.kind = cudaMemcpyHostToDevice;
        CUDA_SAFE_CALL(cudaMemcpy3D(&parm));
        CUDA_CheckError("cudaMemcpy3D");

        // Bind the CUDA memory to a texture reference
        {
                cudaChannelFormatDesc desc = cudaCreateChannelDesc<float4>();
                cudaExtent extent = make_cudaExtent(dimX, dimZ, dimY);
                cudaMalloc3DArray(&wind_vel_array, &desc, extent);

                wind_to_texture_lbm(wind_vel);
                cudaThreadSynchronize();
        }

        free(vel);

	#if PRINT_DEBUG
		printf("Initializing particle distribution functions values\n");
	#endif

        cudaThreadSynchronize();
        
        // Call kernel on GPU
        dim3 block(dimX, 1, 1);
        dim3 grid((dimX*dimY*dimZ) / dimX, 1, 1);

	// Initializes PDFs to a given velocity and rho
        applyInitialValuesGPU_kernel<<<grid, block>>>(obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ, initVel, initRho);

        cudaThreadSynchronize();

        CUDA_CheckError("\napplyInitialBoundariesGPU_kernel\n");
}

/**
 * Main LBM windfield simulation function
 * Makes calls to CUDA kernels: collide, stream, apply boundaries, create windfield
 */
void CUDA_Simulate_WindField_LBM(int collideKernelType, float_type omega, float deltaTime, bool step, int frame, int extra, int boundarySize, float3 boundaryFactor, float_type boundaryRho, float_type force, bool isRendering, float& mlups)
{
        #ifdef PRINT_DEBUG
                printf("Sim windfield: Starting LBM-iteration\n");
        #endif

        double time_collide = 0;
        double time_boundaries = 0;
        double time_stream = 0;
        double time_lbm = 0;
        double time_vel_field = 0;
      
        #if COUNT_MLUPS
                int activeFluidNodes = countNumFluidNodes();
                #if PRINT_DEBUG
                        printf("LBM: Active fluid cells: %d\n", activeFluidNodes);
                #endif
        #endif

        #if PRINT_DEBUG 
                printf("LBM: Starting collide and swap.\n");                       
        #endif
        collideAndSwapGPU(omega, force, extra, time_collide, collideKernelType);
        CUDA_CheckError("collideAndSwapGPU");
        cudaThreadSynchronize();

        #if PRINT_DEBUG
                printf("LBM: Starting stream and swap.\n");
        #endif
        streamAndSwapGPU(extra, time_stream);
        CUDA_CheckError("streamAndSwapGPU");
        cudaThreadSynchronize();
       
        #if COUNT_MLUPS
                mlups = (activeFluidNodes / 1000000.0) / (time_collide + time_stream);
                #if PRINT_DEBUG
                        printf("LBM: Done with collide and stream. MLUPS was: %f\n", mlups);
                #endif
        #endif

        #if PRINT_DEBUG                 
                printf("LBM: Starting outer boundary conditions.\n");
        #endif

        applyOuterBoundariesGPU(extra, time_boundaries, boundarySize, boundaryFactor, boundaryRho);
        CUDA_CheckError("applyOuterBoundariesGPU");
        cudaThreadSynchronize();
 
        #if PRINT_DEBUG
                printf("LBM: Starting to retrieve velocity field.\n");
        #endif       
        createVelocityField(wind_vel, time_vel_field, dimX, dimY, dimZ);

        if(isRendering)
                wind_to_texture_lbm(wind_vel);

        cudaThreadSynchronize();

        #ifdef PRINT_DEBUG
                printf("Sim windfield: Ended LBM-iteration. %d\n");    
        #endif
}

/**
 * Copies the windfield from array to texture on the GPU
 */
void wind_to_texture_lbm(cudaPitchedPtr src) {
        cudaMemcpy3DParms parm = {0};
        parm.srcPtr = src;
        parm.dstArray = wind_vel_array;
        parm.extent = make_cudaExtent(dimX, dimZ, dimY);
        parm.kind = cudaMemcpyDeviceToDevice;
        CUDA_SAFE_CALL(cudaMemcpy3D(&parm));

        wind_vel_tex.filterMode = cudaFilterModeLinear;
        wind_vel_tex.addressMode[0] = cudaAddressModeClamp;
        wind_vel_tex.addressMode[1] = cudaAddressModeClamp;
        wind_vel_tex.addressMode[2] = cudaAddressModeClamp;
        cudaBindTextureToArray(wind_vel_tex, wind_vel_array);
}

/**
 * Free allocated GPU memory, called from the WindLBM class destructor
 */
void CUDA_WindField_Free_LBM() {
	#if PRINT_DEBUG
		printf("LBM sim starting cleanup on the GPU.\n");
	#endif

        free(f0_d);
        free(f1_d);
        free(f2_d);
        free(f3_d);
        free(f4_d);
        free(f5_d);
        free(f6_d);
        free(f7_d);
        free(f8_d);
        free(f9_d);
        free(f10_d);
        free(f11_d);
        free(f12_d);
        free(f13_d);
        free(f14_d);
        free(f15_d);
        free(f16_d);
        free(f17_d);
        free(f18_d);

        CUDA_SAFE_CALL(cudaFree(f0_d));
        CUDA_SAFE_CALL(cudaFree(f1_d));
        CUDA_SAFE_CALL(cudaFree(f2_d));
        CUDA_SAFE_CALL(cudaFree(f3_d));
        CUDA_SAFE_CALL(cudaFree(f4_d));
        CUDA_SAFE_CALL(cudaFree(f5_d));
        CUDA_SAFE_CALL(cudaFree(f6_d));
        CUDA_SAFE_CALL(cudaFree(f7_d));
        CUDA_SAFE_CALL(cudaFree(f8_d));
        CUDA_SAFE_CALL(cudaFree(f9_d));
        CUDA_SAFE_CALL(cudaFree(f10_d));
        CUDA_SAFE_CALL(cudaFree(f11_d));
        CUDA_SAFE_CALL(cudaFree(f12_d));
        CUDA_SAFE_CALL(cudaFree(f13_d));
        CUDA_SAFE_CALL(cudaFree(f14_d));
        CUDA_SAFE_CALL(cudaFree(f15_d));
        CUDA_SAFE_CALL(cudaFree(f16_d));
        CUDA_SAFE_CALL(cudaFree(f17_d));
        CUDA_SAFE_CALL(cudaFree(f18_d));

	CUDA_SAFE_CALL(cudaFree(wind_vel.ptr));
        CUDA_SAFE_CALL(cudaFree(solution.ptr));
        CUDA_SAFE_CALL(cudaFree(obstacle.ptr));
        CUDA_SAFE_CALL(cudaFreeArray(wind_vel_array));

	#if PRINT_DEBUG
		printf("LBM sim finished cleaning up on GPU\n");
	#endif
}

/**
 * Allocate Memory
 */
void allocateMemoryGPU(int size_x, int size_y, int size_z, int boundary_size, float4 velFactor,
float_type *f0_h, float_type *f1_h, float_type *f2_h, float_type *f3_h, float_type *f4_h, float_type *f5_h,
float_type *f6_h, float_type *f7_h, float_type *f8_h, float_type *f9_h, float_type *f10_h, float_type *f11_h,
float_type *f12_h, float_type *f13_h, float_type *f14_h, float_type *f15_h, float_type *f16_h, float_type *f17_h, float_type *f18_h)
{
        // Used for transfer to constant memory
        int *temp_lbm_ysize = (int *)malloc(sizeof(int));
        int *temp_lbm_pitch = (int *)malloc(sizeof(int));
	int *temp_lbm_boundary_size = (int *)malloc(sizeof(int));
	float4 *temp_lbm_vel_factor = (float4 *)malloc(sizeof(float4));

        dimX = size_x;
        dimY = size_y;
        dimZ = size_z;

        // The destination pointers for for the particle distribution functions
        f_d[0] = make_cudaPitchedPtr(f0_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[1] = make_cudaPitchedPtr(f1_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[2] = make_cudaPitchedPtr(f2_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[3] = make_cudaPitchedPtr(f3_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[4] = make_cudaPitchedPtr(f4_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[5] = make_cudaPitchedPtr(f5_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[6] = make_cudaPitchedPtr(f6_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[7] = make_cudaPitchedPtr(f7_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[8] = make_cudaPitchedPtr(f8_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[9] = make_cudaPitchedPtr(f9_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[10] = make_cudaPitchedPtr(f10_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[11] = make_cudaPitchedPtr(f11_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[12] = make_cudaPitchedPtr(f12_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[13] = make_cudaPitchedPtr(f13_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[14] = make_cudaPitchedPtr(f14_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[15] = make_cudaPitchedPtr(f15_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[16] = make_cudaPitchedPtr(f16_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[17] = make_cudaPitchedPtr(f17_d, dimX * sizeof(float_type), dimX, dimY );
        f_d[18] = make_cudaPitchedPtr(f18_d, dimX * sizeof(float_type), dimX, dimY );
        
        cudaExtent f_extent = make_cudaExtent(dimX * sizeof(float_type), dimY, dimZ);

        // Allocate device memory for the particle distribution functions
        for(int i=0; i<19; i++)
        {
                CUDA_SAFE_CALL(cudaMalloc3D(&f_d[i], f_extent));
                CUDA_CheckError("Particle distribution functions cudaMalloc3D");
        }
        
        // Copy the particle distribution functions from the host memory to device memory
        cudaMemcpy3DParms p = {0};
        p.extent = f_extent;
        p.kind = cudaMemcpyHostToDevice;

        // The source
        p.srcPtr = make_cudaPitchedPtr(f0_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[0];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[0] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f1_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[1];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[1] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f2_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[2];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[2] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f3_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[3];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[3] cudaMemcpy3D");
        
        // The source
        p.srcPtr = make_cudaPitchedPtr(f4_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[4];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[4] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f5_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[5];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[5] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f6_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[6];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[6] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f7_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[7];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[7] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f8_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[8];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[8] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f9_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[9];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[9] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f10_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[10];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[10] cudaMemcpy3D");
        // The source
        p.srcPtr = make_cudaPitchedPtr(f11_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[11];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[11] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f12_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[12];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[12] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f13_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[13];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[13] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f14_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[14];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[14] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f15_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[15];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[15] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f16_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[16];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[16] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f17_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[17];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[17] cudaMemcpy3D");

        // The source
        p.srcPtr = make_cudaPitchedPtr(f18_h, dimX * sizeof(float_type), dimX, dimY );
        // The destination
        p.dstPtr = f_d[18];
        CUDA_SAFE_CALL(cudaMemcpy3D(&p));
        CUDA_CheckError("f_d[18] cudaMemcpy3D");

        for(int i=0; i<19; i++)
        {
                pdf.ptr[i] = f_d[i].ptr;
        }

        // Transfer convenience variables to constant memory
        // (to avoid too many kernel arguments)
        *temp_lbm_ysize = f_d[0].ysize;
        *temp_lbm_pitch = f_d[0].pitch;
	*temp_lbm_boundary_size = boundary_size;
	*temp_lbm_vel_factor = velFactor;

	cudaMemcpyToSymbol(lbm_vel_factor_dev, temp_lbm_vel_factor, sizeof(float4));
        cudaMemcpyToSymbol(lbm_ysize_dev, temp_lbm_ysize, sizeof(int));
        cudaMemcpyToSymbol(lbm_pitch_dev, temp_lbm_pitch, sizeof(int));
        cudaMemcpyToSymbol(lbm_boundary_size_dev, temp_lbm_boundary_size, sizeof(int));
}

/**
 * Retrieve a velocityfield by calculating macroscopic velocity for all the lattice nodes
 */
void createVelocityField(cudaPitchedPtr vel_field, double &time, const int dimX, const int dimY, const int dimZ) {
        #if PRINT_DEBUG
		printf("Starting createVelocityField, xdim: %d\n", dimX);
	#endif
        //dim3 blockSize(dimX);
        //dim3 nBlocks(dimY,dimZ);

        cudaThreadSynchronize();

        unsigned int timer = 0;
        CUT_SAFE_CALL(cutCreateTimer(&timer));
        CUT_SAFE_CALL(cutStartTimer(timer));

        //createVelocityField_Kernel<<<nBlocks, blockSize>>>(obstacle, vel_field, pdf, dimX, dimY, dimZ);

		//setting cache preferance
		//cudaFuncSetCacheConfig(createVelocityField_Kernel2, cudaFuncCachePreferL1);

        dim3 blockSize(dimX);
        dim3 nBlocks(dimY, dimZ);
        createVelocityField_Kernel2<<<nBlocks, blockSize>>>(obstacle, vel_field, pdf, dimX, dimY, dimZ);

        cudaThreadSynchronize();

	#if PRINT_DEBUG
	        printf("Ended create velocity field kernel.\n\n");
	#endif
        CUT_SAFE_CALL(cutStopTimer(timer));
        time = cutGetTimerValue(timer)/1000;
        CUT_SAFE_CALL(cutDeleteTimer(timer));

        CUDA_CheckError("createVelocityField_Kernel2\n");

	#if PRINT_DEBUG
	        printf("Returning from velocity field kernel\n");
	#endif
}

/**
 * Apply outer boundary conditions
 */
void applyOuterBoundariesGPU(int extra, double &time, int boundarySize, float3 boundaryFactor, float_type boundaryRho)
{
        #if PRINT_DEBUG
		printf("Apply outer boundaries: solid size %d\n", obstacle.xsize * obstacle.ysize);
	        printf("Apply outer boundaries: dim size %d\n", dimX * dimY * dimZ);
	#endif

        cudaThreadSynchronize();

        unsigned int timer = 0;
        CUT_SAFE_CALL(cutCreateTimer(&timer));

        // Old boundary kernels, run on all cells, useful for debugging
        /*dim3 block(dimX, 1, 1);
        dim3 grid((dimX*dimY*dimZ) / dimX, 1, 1);
        CUT_SAFE_CALL(cutStartTimer(timer));
        applyOuterBoundariesGPU_kernel<<<grid, block>>>(obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ, boundaryFactor, initRho);*/

	// Run if outflow on X = dimX-1 is desired
        /*dim3 block(dimY, 1, 1);
        dim3 grid(dimZ, boundarySize, 1);
	applyOutFlowGPU_kernel<<<grid, block>>>(obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ);*/

	// Multiply boundaryvelocity for "safe" benchmarking across different collide kernels
	//boundaryFactor.x *= 0.01; boundaryFactor.y *= 0.01; boundaryFactor.z *= 0.01;

	//setting cache preferance
	//cudaFuncSetCacheConfig(applyOuterBoundariesGPU_kernelYZ, cudaFuncCachePreferL1);
	//cudaFuncSetCacheConfig(applyOuterBoundariesGPU_kernelXY, cudaFuncCachePreferL1);
	//cudaFuncSetCacheConfig(applyOuterBoundariesGPU_kernelXZ, cudaFuncCachePreferL1);


	// Inflow boundaries along all the walls
	dim3 block1(dimY, 1, 1);
        dim3 grid1(boundarySize, dimZ*2, 1);
        CUT_SAFE_CALL(cutStartTimer(timer));
	applyOuterBoundariesGPU_kernelYZ<<<grid1, block1>>>(obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ, boundaryFactor, boundaryRho);
	dim3 block2(dimX, 1, 1);
        dim3 grid2(dimY*2, boundarySize, 1);
	applyOuterBoundariesGPU_kernelXY<<<grid2, block2>>>(obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ, boundaryFactor, boundaryRho);
	dim3 block3(dimX, 1, 1);
        dim3 grid3(dimZ*2, boundarySize, 1);
	applyOuterBoundariesGPU_kernelXZ<<<grid3, block3>>>(obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ, boundaryFactor, boundaryRho);
        cudaThreadSynchronize();

        CUT_SAFE_CALL(cutStopTimer(timer));
        time = cutGetTimerValue(timer)/1000;
        CUT_SAFE_CALL(cutDeleteTimer(timer));

        CUDA_CheckError("\napplyOuterBoundariesGPU_kernel\n");
}

/**
 * Count the number of active fluid nodes (not obstacles) in GPU memory
 * Copies the whole buffer from GPU to CPU for the reduction
 * Should be implemented in a kernel, this is only for benchmarking
 */
int countNumFluidNodes()
{
        dim3 blockSize(dimX);
        dim3 nBlocks(dimY,dimZ);
	int numCells = 0;

        cudaThreadSynchronize();

	int* h_obs = (int*)malloc(dim.x*dim.z*dim.y*sizeof(int));

	obstacle_extent = make_cudaExtent(dim.x*sizeof(int), dim.z, dim.y);
        cudaMemcpy3DParms obs_params = {0};
	obs_params.srcPtr.ptr = obstacle.ptr;
	obs_params.srcPtr.pitch = obstacle.pitch;
	obs_params.dstPtr.ptr = h_obs;
	obs_params.extent = obstacle_extent;
	obs_params.kind = cudaMemcpyDeviceToHost;

	cudaMemcpy3D(&obs_params);

	for(int i=0;i<dim.x*dim.z*dim.y;i++) {
		if(h_obs[i] == 0)
			numCells++;
	}

	free(h_obs);

        CUDA_CheckError("\ncountNumFluidNodes\n");

	return numCells;
}

/**
 * Stream And Swap
 */
void streamAndSwapGPU(int extra, double &time)
{
        dim3 blockSize(dimX);
        dim3 nBlocks(dimY,dimZ);

        cudaThreadSynchronize();

        unsigned int timer = 0;
        CUT_SAFE_CALL(cutCreateTimer(&timer));
        CUT_SAFE_CALL(cutStartTimer(timer));

		//setting cache preferance
#ifdef FERMI_OPT
		cudaFuncSetCacheConfig(streamAndSwapGPU_kernel, cudaFuncCachePreferL1);
#endif


        streamAndSwapGPU_kernel<<<nBlocks, blockSize>>>(obstacle, pdf, f_d[0].pitch, dimX, dimY, dimZ, extra);

        cudaThreadSynchronize();

        CUT_SAFE_CALL(cutStopTimer(timer));

	time = cutGetTimerValue(timer)/1000;
        CUT_SAFE_CALL(cutDeleteTimer(timer));

        CUDA_CheckError("\nstreamAndSwapGPU_kernel\n");
}

/**
 * Collide And Swap 
 */
void collideAndSwapGPU(float_type omega, float_type force, int extra, double &time, int kernel_type)
{
        cudaThreadSynchronize();

        unsigned int timer = 0;
        CUT_SAFE_CALL(cutCreateTimer(&timer));

        if(kernel_type == 0)
        {
                dim3 blockSize(dimX);
                dim3 nBlocks(dimY, dimZ);
		#if PRINT_DEBUG
			printf("LBM-GPU: Running precise collide and swap kernel.\n");
		#endif
                CUT_SAFE_CALL(cutStartTimer(timer));
                collideAndSwapGPU_RegularBad<<<nBlocks, blockSize>>>(omega, force, obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ);
                cudaThreadSynchronize();
                CUT_SAFE_CALL(cutStopTimer(timer));
				
		#if PRINT_DEBUG
			printf("LBM-GPU: Returned from precise collide and swap kernel.\n");
		#endif
        }
	
	if(kernel_type == 1)
        {
                dim3 blockSize(dimY);
                dim3 nBlocks(dimX, dimZ);
		
                #if PRINT_DEBUG
			printf("LBM-GPU: Running shared memory collide and swap kernel.\n");
		#endif
                CUT_SAFE_CALL(cutStartTimer(timer));
				#ifdef FERMI_OPT
				//setting cache preferance
				cudaFuncSetCacheConfig(collideAndSwapGPU_kernelShared3, cudaFuncCachePreferShared);
				#endif
                collideAndSwapGPU_kernelShared3<<<nBlocks, blockSize, dimY * sizeof(float)*19>>>(omega, force, obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ);
				//collideAndSwapGPU_kernelShared4<<<nBlocks, blockSize>>>(omega, force, obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ);
                cudaThreadSynchronize();
                CUT_SAFE_CALL(cutStopTimer(timer));
		#if PRINT_DEBUG
			printf("LBM-GPU: Returned from shared memory collide and swap kernel.\n");
		#endif
        }

        if(kernel_type == 2)
        {
                dim3 blockSize(dimX);
                dim3 nBlocks(dimY, dimZ);

		#if PRINT_DEBUG
	                printf("LBM-GPU: Running precise collide and swap kernel.\n");
		#endif
				#ifdef FERMI_OPT
					//setting cache preferance
					cudaFuncSetCacheConfig(collideAndSwapGPU_kernelPrecise2, cudaFuncCachePreferL1);
				#endif
                CUT_SAFE_CALL(cutStartTimer(timer));
                collideAndSwapGPU_kernelPrecise2<<<nBlocks, blockSize>>>(omega, force, obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ);
                cudaThreadSynchronize();
                CUT_SAFE_CALL(cutStopTimer(timer));
				
		#if PRINT_DEBUG
	                printf("LBM-GPU: Returned from precise collide and swap kernel.\n");
		#endif
        }
 	
	if(kernel_type == 3)
        {
                dim3 blockSize(dimX);
                dim3 nBlocks(dimY, dimZ);

		#if PRINT_DEBUG
	                printf("LBM-GPU: Running double precision collide and swap kernel.\n");
		#endif
				#ifdef FERMI_OPT
				//setting cache preferance
				cudaFuncSetCacheConfig(collideAndSwapGPU_kernel5Double, cudaFuncCachePreferL1);
				#endif
                CUT_SAFE_CALL(cutStartTimer(timer));
                collideAndSwapGPU_kernel5Double<<<nBlocks, blockSize>>>(omega, force, obstacle, pdf, f_d[0].pitch, f_d[0].ysize, dimX, dimY, dimZ);
                cudaThreadSynchronize();
                CUT_SAFE_CALL(cutStopTimer(timer));
		#if PRINT_DEBUG
	                printf("LBM-GPU: Returned from double precision collide and swap kernel.\n");
		#endif
        }

        time = cutGetTimerValue(timer)/1000;
        CUT_SAFE_CALL(cutDeleteTimer(timer));
        cutDeleteTimer(timer);
		//printf("collide time: %f \n",time);

        CUDA_CheckError("\ncollideAndSwapGPU_kernel\n");
	#if PRINT_DEBUG
	        printf("LBM_GPU: Returning from collide and swap function.\n");
	#endif
}
